# -*- coding: utf-8 -*-
#---------------------------------------------------------------------------
# Copyright 2023 VMware, Inc.  All rights reserved.

# AUTO GENERATED FILE -- DO NOT MODIFY!
#
# vAPI stub file for package com.vmware.nsx.dns.forwarders.
#---------------------------------------------------------------------------

"""


"""

__author__ = 'VMware, Inc.'
__docformat__ = 'restructuredtext en'

import sys
from warnings import warn

from vmware.vapi.bindings import type
from vmware.vapi.bindings.converter import TypeConverter
from vmware.vapi.bindings.enum import Enum
from vmware.vapi.bindings.error import VapiError
from vmware.vapi.bindings.struct import VapiStruct
from vmware.vapi.bindings.stub import (
    ApiInterfaceStub, StubFactoryBase, VapiInterface)
from vmware.vapi.bindings.common import raise_core_exception
from vmware.vapi.data.validator import (UnionValidator, HasFieldsOfValidator)
from vmware.vapi.exception import CoreException
from vmware.vapi.lib.constants import TaskType
from vmware.vapi.lib.rest import OperationRestMetadata


class FailedQueries(VapiInterface):
    """
    
    """

    _VAPI_SERVICE_ID = 'com.vmware.nsx.dns.forwarders.failed_queries'
    """
    Identifier of the service in canonical form.
    """
    def __init__(self, config):
        """
        :type  config: :class:`vmware.vapi.bindings.stub.StubConfiguration`
        :param config: Configuration to be used for creating the stub.
        """
        VapiInterface.__init__(self, config, _FailedQueriesStub)
        self._VAPI_OPERATION_IDS = {}


    def get(self,
            forwarder_id,
            count=None,
            ):
        """
        Return the given count of recent failed DNS queries from DNS forwarder.
        Since the DNS forwarder is running in Acitve/Standby HA mode on
        transport nodes, the given count of queries will be returned from each
        nodes. Hence the total queries returned could be doubled. If no count
        is specified, 100 recent failed queries are returned. If the recent
        failures is less than the given count, all the failures will be
        returned. The maximum count is 1,000.

        :type  forwarder_id: :class:`str`
        :param forwarder_id: (required)
        :type  count: :class:`long` or ``None``
        :param count: The count of the failed DNS queries (optional, default to 100)
        :rtype: :class:`com.vmware.nsx.model_client.DnsFailedQueries`
        :return: com.vmware.nsx.model.DnsFailedQueries
        :raise: :class:`com.vmware.vapi.std.errors_client.InvalidRequest` 
            Bad Request, Precondition Failed
        :raise: :class:`com.vmware.vapi.std.errors_client.Unauthorized` 
            Forbidden
        :raise: :class:`com.vmware.vapi.std.errors_client.ServiceUnavailable` 
            Service Unavailable
        :raise: :class:`com.vmware.vapi.std.errors_client.InternalServerError` 
            Internal Server Error
        :raise: :class:`com.vmware.vapi.std.errors_client.NotFound` 
            Not Found
        """
        return self._invoke('get',
                            {
                            'forwarder_id': forwarder_id,
                            'count': count,
                            })
class Nslookup(VapiInterface):
    """
    
    """

    _VAPI_SERVICE_ID = 'com.vmware.nsx.dns.forwarders.nslookup'
    """
    Identifier of the service in canonical form.
    """
    def __init__(self, config):
        """
        :type  config: :class:`vmware.vapi.bindings.stub.StubConfiguration`
        :param config: Configuration to be used for creating the stub.
        """
        VapiInterface.__init__(self, config, _NslookupStub)
        self._VAPI_OPERATION_IDS = {}


    def get(self,
            forwarder_id,
            address=None,
            server_ip=None,
            source_ip=None,
            ):
        """
        Query the nameserver for an ip-address or a FQDN of the given an
        address optionally using an specified DNS server. If the address is a
        fqdn, nslookup will resolve ip-address with it. If the address is an
        ip-address, do a reverse lookup and answer fqdn(s).

        .. deprecated:: unknown

        :type  forwarder_id: :class:`str`
        :param forwarder_id: (required)
        :type  address: :class:`str` or ``None``
        :param address: IP address or FQDN for nslookup (optional)
        :type  server_ip: :class:`str` or ``None``
        :param server_ip: IPv4 address (optional)
        :type  source_ip: :class:`str` or ``None``
        :param source_ip: IPv4 address (optional)
        :rtype: :class:`com.vmware.nsx.model_client.DnsAnswer`
        :return: com.vmware.nsx.model.DnsAnswer
        :raise: :class:`com.vmware.vapi.std.errors_client.InvalidRequest` 
            Bad Request, Precondition Failed
        :raise: :class:`com.vmware.vapi.std.errors_client.Unauthorized` 
            Forbidden
        :raise: :class:`com.vmware.vapi.std.errors_client.ServiceUnavailable` 
            Service Unavailable
        :raise: :class:`com.vmware.vapi.std.errors_client.InternalServerError` 
            Internal Server Error
        :raise: :class:`com.vmware.vapi.std.errors_client.NotFound` 
            Not Found
        """
        warn('com.vmware.nsx.dns.forwarders.Nslookup.get is deprecated.', DeprecationWarning)
        return self._invoke('get',
                            {
                            'forwarder_id': forwarder_id,
                            'address': address,
                            'server_ip': server_ip,
                            'source_ip': source_ip,
                            })
class State(VapiInterface):
    """
    
    """

    _VAPI_SERVICE_ID = 'com.vmware.nsx.dns.forwarders.state'
    """
    Identifier of the service in canonical form.
    """
    def __init__(self, config):
        """
        :type  config: :class:`vmware.vapi.bindings.stub.StubConfiguration`
        :param config: Configuration to be used for creating the stub.
        """
        VapiInterface.__init__(self, config, _StateStub)
        self._VAPI_OPERATION_IDS = {}


    def get(self,
            forwarder_id,
            barrier_id=None,
            request_id=None,
            ):
        """
        Return the realized state information of a DNS forwarder. After a DNS
        forwarder was created or updated, you can invoke this API to check the
        realization state of the forwarder.

        .. deprecated:: unknown

        :type  forwarder_id: :class:`str`
        :param forwarder_id: (required)
        :type  barrier_id: :class:`long` or ``None``
        :param barrier_id: (optional)
        :type  request_id: :class:`str` or ``None``
        :param request_id: Realization request ID (optional)
        :rtype: :class:`com.vmware.nsx.model_client.ConfigurationState`
        :return: com.vmware.nsx.model.ConfigurationState
        :raise: :class:`com.vmware.vapi.std.errors_client.InvalidRequest` 
            Bad Request, Precondition Failed
        :raise: :class:`com.vmware.vapi.std.errors_client.Unauthorized` 
            Forbidden
        :raise: :class:`com.vmware.vapi.std.errors_client.ServiceUnavailable` 
            Service Unavailable
        :raise: :class:`com.vmware.vapi.std.errors_client.InternalServerError` 
            Internal Server Error
        :raise: :class:`com.vmware.vapi.std.errors_client.NotFound` 
            Not Found
        """
        warn('com.vmware.nsx.dns.forwarders.State.get is deprecated.', DeprecationWarning)
        return self._invoke('get',
                            {
                            'forwarder_id': forwarder_id,
                            'barrier_id': barrier_id,
                            'request_id': request_id,
                            })
class Statistics(VapiInterface):
    """
    
    """

    _VAPI_SERVICE_ID = 'com.vmware.nsx.dns.forwarders.statistics'
    """
    Identifier of the service in canonical form.
    """
    def __init__(self, config):
        """
        :type  config: :class:`vmware.vapi.bindings.stub.StubConfiguration`
        :param config: Configuration to be used for creating the stub.
        """
        VapiInterface.__init__(self, config, _StatisticsStub)
        self._VAPI_OPERATION_IDS = {}


    def get(self,
            forwarder_id,
            ):
        """
        Returns the statistics of the given dns forwarder specified by
        forwarder id.

        .. deprecated:: unknown

        :type  forwarder_id: :class:`str`
        :param forwarder_id: (required)
        :rtype: :class:`com.vmware.nsx.model_client.DnsForwarderStatistics`
        :return: com.vmware.nsx.model.DnsForwarderStatistics
        :raise: :class:`com.vmware.vapi.std.errors_client.InvalidRequest` 
            Bad Request, Precondition Failed
        :raise: :class:`com.vmware.vapi.std.errors_client.Unauthorized` 
            Forbidden
        :raise: :class:`com.vmware.vapi.std.errors_client.ServiceUnavailable` 
            Service Unavailable
        :raise: :class:`com.vmware.vapi.std.errors_client.InternalServerError` 
            Internal Server Error
        :raise: :class:`com.vmware.vapi.std.errors_client.NotFound` 
            Not Found
        """
        warn('com.vmware.nsx.dns.forwarders.Statistics.get is deprecated.', DeprecationWarning)
        return self._invoke('get',
                            {
                            'forwarder_id': forwarder_id,
                            })
class Status(VapiInterface):
    """
    
    """

    _VAPI_SERVICE_ID = 'com.vmware.nsx.dns.forwarders.status'
    """
    Identifier of the service in canonical form.
    """
    def __init__(self, config):
        """
        :type  config: :class:`vmware.vapi.bindings.stub.StubConfiguration`
        :param config: Configuration to be used for creating the stub.
        """
        VapiInterface.__init__(self, config, _StatusStub)
        self._VAPI_OPERATION_IDS = {}


    def get(self,
            forwarder_id,
            ):
        """
        Returns the current status of the given DNS forwarder.

        .. deprecated:: unknown

        :type  forwarder_id: :class:`str`
        :param forwarder_id: (required)
        :rtype: :class:`com.vmware.nsx.model_client.DnsForwarderStatus`
        :return: com.vmware.nsx.model.DnsForwarderStatus
        :raise: :class:`com.vmware.vapi.std.errors_client.InvalidRequest` 
            Bad Request, Precondition Failed
        :raise: :class:`com.vmware.vapi.std.errors_client.Unauthorized` 
            Forbidden
        :raise: :class:`com.vmware.vapi.std.errors_client.ServiceUnavailable` 
            Service Unavailable
        :raise: :class:`com.vmware.vapi.std.errors_client.InternalServerError` 
            Internal Server Error
        :raise: :class:`com.vmware.vapi.std.errors_client.NotFound` 
            Not Found
        """
        warn('com.vmware.nsx.dns.forwarders.Status.get is deprecated.', DeprecationWarning)
        return self._invoke('get',
                            {
                            'forwarder_id': forwarder_id,
                            })
class _FailedQueriesStub(ApiInterfaceStub):
    def __init__(self, config):
        # properties for get operation
        get_input_type = type.StructType('operation-input', {
            'forwarder_id': type.StringType(),
            'count': type.OptionalType(type.IntegerType()),
        })
        get_error_dict = {
            'com.vmware.vapi.std.errors.invalid_request':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'InvalidRequest'),
            'com.vmware.vapi.std.errors.unauthorized':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'Unauthorized'),
            'com.vmware.vapi.std.errors.service_unavailable':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'ServiceUnavailable'),
            'com.vmware.vapi.std.errors.internal_server_error':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'InternalServerError'),
            'com.vmware.vapi.std.errors.not_found':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'NotFound'),

        }
        get_input_value_validator_list = [
        ]
        get_output_validator_list = [
        ]
        get_rest_metadata = OperationRestMetadata(
            http_method='GET',
            url_template='/api/v1/dns/forwarders/{forwarder-id}/failed-queries',
            path_variables={
                'forwarder_id': 'forwarder-id',
            },
             header_parameters={
                 },
            query_parameters={
                'count': 'count',
            },
            content_type='application/json'
        )

        operations = {
            'get': {
                'input_type': get_input_type,
                'output_type': type.ReferenceType('com.vmware.nsx.model_client', 'DnsFailedQueries'),
                'errors': get_error_dict,
                'input_value_validator_list': get_input_value_validator_list,
                'output_validator_list': get_output_validator_list,
                'task_type': TaskType.NONE,
            },
        }
        rest_metadata = {
            'get': get_rest_metadata,
        }
        ApiInterfaceStub.__init__(
            self, iface_name='com.vmware.nsx.dns.forwarders.failed_queries',
            config=config, operations=operations, rest_metadata=rest_metadata,
            is_vapi_rest=False)

class _NslookupStub(ApiInterfaceStub):
    def __init__(self, config):
        # properties for get operation
        get_input_type = type.StructType('operation-input', {
            'forwarder_id': type.StringType(),
            'address': type.OptionalType(type.StringType()),
            'server_ip': type.OptionalType(type.StringType()),
            'source_ip': type.OptionalType(type.StringType()),
        })
        get_error_dict = {
            'com.vmware.vapi.std.errors.invalid_request':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'InvalidRequest'),
            'com.vmware.vapi.std.errors.unauthorized':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'Unauthorized'),
            'com.vmware.vapi.std.errors.service_unavailable':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'ServiceUnavailable'),
            'com.vmware.vapi.std.errors.internal_server_error':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'InternalServerError'),
            'com.vmware.vapi.std.errors.not_found':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'NotFound'),

        }
        get_input_value_validator_list = [
        ]
        get_output_validator_list = [
        ]
        get_rest_metadata = OperationRestMetadata(
            http_method='GET',
            url_template='/api/v1/dns/forwarders/{forwarder-id}/nslookup',
            path_variables={
                'forwarder_id': 'forwarder-id',
            },
             header_parameters={
                     },
            query_parameters={
                'address': 'address',
                'server_ip': 'server_ip',
                'source_ip': 'source_ip',
            },
            content_type='application/json'
        )

        operations = {
            'get': {
                'input_type': get_input_type,
                'output_type': type.ReferenceType('com.vmware.nsx.model_client', 'DnsAnswer'),
                'errors': get_error_dict,
                'input_value_validator_list': get_input_value_validator_list,
                'output_validator_list': get_output_validator_list,
                'task_type': TaskType.NONE,
            },
        }
        rest_metadata = {
            'get': get_rest_metadata,
        }
        ApiInterfaceStub.__init__(
            self, iface_name='com.vmware.nsx.dns.forwarders.nslookup',
            config=config, operations=operations, rest_metadata=rest_metadata,
            is_vapi_rest=False)

class _StateStub(ApiInterfaceStub):
    def __init__(self, config):
        # properties for get operation
        get_input_type = type.StructType('operation-input', {
            'forwarder_id': type.StringType(),
            'barrier_id': type.OptionalType(type.IntegerType()),
            'request_id': type.OptionalType(type.StringType()),
        })
        get_error_dict = {
            'com.vmware.vapi.std.errors.invalid_request':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'InvalidRequest'),
            'com.vmware.vapi.std.errors.unauthorized':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'Unauthorized'),
            'com.vmware.vapi.std.errors.service_unavailable':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'ServiceUnavailable'),
            'com.vmware.vapi.std.errors.internal_server_error':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'InternalServerError'),
            'com.vmware.vapi.std.errors.not_found':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'NotFound'),

        }
        get_input_value_validator_list = [
        ]
        get_output_validator_list = [
        ]
        get_rest_metadata = OperationRestMetadata(
            http_method='GET',
            url_template='/api/v1/dns/forwarders/{forwarder-id}/state',
            path_variables={
                'forwarder_id': 'forwarder-id',
            },
             header_parameters={
                   },
            query_parameters={
                'barrier_id': 'barrier_id',
                'request_id': 'request_id',
            },
            content_type='application/json'
        )

        operations = {
            'get': {
                'input_type': get_input_type,
                'output_type': type.ReferenceType('com.vmware.nsx.model_client', 'ConfigurationState'),
                'errors': get_error_dict,
                'input_value_validator_list': get_input_value_validator_list,
                'output_validator_list': get_output_validator_list,
                'task_type': TaskType.NONE,
            },
        }
        rest_metadata = {
            'get': get_rest_metadata,
        }
        ApiInterfaceStub.__init__(
            self, iface_name='com.vmware.nsx.dns.forwarders.state',
            config=config, operations=operations, rest_metadata=rest_metadata,
            is_vapi_rest=False)

class _StatisticsStub(ApiInterfaceStub):
    def __init__(self, config):
        # properties for get operation
        get_input_type = type.StructType('operation-input', {
            'forwarder_id': type.StringType(),
        })
        get_error_dict = {
            'com.vmware.vapi.std.errors.invalid_request':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'InvalidRequest'),
            'com.vmware.vapi.std.errors.unauthorized':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'Unauthorized'),
            'com.vmware.vapi.std.errors.service_unavailable':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'ServiceUnavailable'),
            'com.vmware.vapi.std.errors.internal_server_error':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'InternalServerError'),
            'com.vmware.vapi.std.errors.not_found':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'NotFound'),

        }
        get_input_value_validator_list = [
        ]
        get_output_validator_list = [
        ]
        get_rest_metadata = OperationRestMetadata(
            http_method='GET',
            url_template='/api/v1/dns/forwarders/{forwarder-id}/statistics',
            path_variables={
                'forwarder_id': 'forwarder-id',
            },
             header_parameters={
               },
            query_parameters={
            },
            content_type='application/json'
        )

        operations = {
            'get': {
                'input_type': get_input_type,
                'output_type': type.ReferenceType('com.vmware.nsx.model_client', 'DnsForwarderStatistics'),
                'errors': get_error_dict,
                'input_value_validator_list': get_input_value_validator_list,
                'output_validator_list': get_output_validator_list,
                'task_type': TaskType.NONE,
            },
        }
        rest_metadata = {
            'get': get_rest_metadata,
        }
        ApiInterfaceStub.__init__(
            self, iface_name='com.vmware.nsx.dns.forwarders.statistics',
            config=config, operations=operations, rest_metadata=rest_metadata,
            is_vapi_rest=False)

class _StatusStub(ApiInterfaceStub):
    def __init__(self, config):
        # properties for get operation
        get_input_type = type.StructType('operation-input', {
            'forwarder_id': type.StringType(),
        })
        get_error_dict = {
            'com.vmware.vapi.std.errors.invalid_request':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'InvalidRequest'),
            'com.vmware.vapi.std.errors.unauthorized':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'Unauthorized'),
            'com.vmware.vapi.std.errors.service_unavailable':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'ServiceUnavailable'),
            'com.vmware.vapi.std.errors.internal_server_error':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'InternalServerError'),
            'com.vmware.vapi.std.errors.not_found':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'NotFound'),

        }
        get_input_value_validator_list = [
        ]
        get_output_validator_list = [
        ]
        get_rest_metadata = OperationRestMetadata(
            http_method='GET',
            url_template='/api/v1/dns/forwarders/{forwarder-id}/status',
            path_variables={
                'forwarder_id': 'forwarder-id',
            },
             header_parameters={
               },
            query_parameters={
            },
            content_type='application/json'
        )

        operations = {
            'get': {
                'input_type': get_input_type,
                'output_type': type.ReferenceType('com.vmware.nsx.model_client', 'DnsForwarderStatus'),
                'errors': get_error_dict,
                'input_value_validator_list': get_input_value_validator_list,
                'output_validator_list': get_output_validator_list,
                'task_type': TaskType.NONE,
            },
        }
        rest_metadata = {
            'get': get_rest_metadata,
        }
        ApiInterfaceStub.__init__(
            self, iface_name='com.vmware.nsx.dns.forwarders.status',
            config=config, operations=operations, rest_metadata=rest_metadata,
            is_vapi_rest=False)


class StubFactory(StubFactoryBase):
    _attrs = {
        'FailedQueries': FailedQueries,
        'Nslookup': Nslookup,
        'State': State,
        'Statistics': Statistics,
        'Status': Status,
    }




"""
Twisted server
"""

__author__ = 'VMware, Inc.'
__copyright__ = 'Copyright 2015, 2017, 2021 VMware, Inc.  All rights reserved. -- VMware Confidential'  # pylint: disable=line-too-long

import threading
import six
import warnings

import twisted
# Without adding this explicit import, the ssl part of twisted
# doesn't get initialized
import twisted.internet.ssl
from twisted.internet import defer
from twisted.internet.protocol import Protocol, Factory
from twisted.web.resource import Resource
from twisted.web.server import Site

from vmware.vapi.server.server_interface import ServerInterface
from vmware.vapi.lib.addr_url_parser import parse_addr_url
from vmware.vapi.lib.log import get_vapi_logger

# Try to use epoll if available
try:
    twisted.internet.epollreactor.install()
except Exception:
    pass

# Twisted log
#import sys
#twisted.python.log.startLogging(sys.stdout, setStdout=False)

# NYI: Debug
# import time

logger = get_vapi_logger(__name__)

# Singleton twisted server
_server_lock = threading.Lock()
_twisted_server = None


class TwistedConnectionAdapter(object):
    """ Twisted connection adapter """

    def __init__(self, transport):
        """
        Twisted connection adapter init

        :type  transport: :class:`twisted.internet.interfaces.ITransport`
        :param transport: Twisted transport
        """
        self.transport = transport

    def write(self, data):
        """
        write

        :type  data: :class:`str`
        :param data: data to write
        """
        # Only need to do this for async protocol
        self.transport.reactor.callFromThread(self._write, data)

    def close(self):
        """ close """
        # Only need to do this for async protocol
        self.transport.reactor.callFromThread(self._close)

    def _write(self, data):
        """
        Internal write

        :type  data: :class:`str`
        :param data: data to write
        """
        self.transport.write(data)

    def _close(self):
        """ Internal close """
        self.transport.loseConnection()
        self.transport = None


class TwistedAsyncProtocolHandlerAdapter(Protocol):  # pylint: disable=W0232
    """ Twisted async protocol handler adapter """

    def connectionMade(self):
        """ connection established callback """
        connection = TwistedConnectionAdapter(self.transport)
        self.data_handler = self.factory.protocol_handler.get_data_handler(
            connection)

    def dataReceived(self, data):
        """
        data received

        :type  data: :class:`str`
        :param data: data to write
        """
        self.data_handler.data_ready(data)

    def connectionLost(self, reason):  # pylint: disable=W0222
        """
        Twisted connection lost

        :type  reason: :class:`str`
        :param reason: Connection lost reason
        """
        if issubclass(twisted.internet.error.ConnectionDone, reason.type):
            logger.debug('connectionLost: reason %s', reason)
            self.data_handler.data_end()
        else:
            logger.error('connectionLost: reason %s', reason)
            self.data_handler.data_abort()
        self.data_handler = None  # pylint: disable=W0201


class TwistedVmwareProtocolFactory(Factory):
    """ Twisted vmacre protocol factory """
    protocol = TwistedAsyncProtocolHandlerAdapter

    def __init__(self, protocol_handler):
        """
        Twisted vmacre protocol factory init

        :type  protocol_handler: :class:`vmware.vapi.protocol.server.transport.\
            async_protocol_handler.AsyncProtocolHandler`
        :param protocol_handler: protocol handler for this addr
        """
        self.protocol_handler = protocol_handler


class TwistedHttpConnectionAdapter(object):
    """ Twisted http connection adapter """

    def __init__(self, request):
        """
        Twisted http connection init

        :type  request: :class:`twisted.web.http.Request`
        :param request: Twisted http request object
        """
        self.request = request

    def write(self, data):
        """
        write

        :type  data: :class:`str`
        :param data: data to write
        """
        self.request.transport.reactor.callFromThread(self._write, data)

    def close(self):
        """ close """
        self.request.transport.reactor.callFromThread(self._close)

    # Connection interface
    def _write(self, data):
        """ Internal write """
        if not self.request.finished:
            # NYI: Debug only
            # logger.debug('%f: write data %d', time.time(), len(data))
            self.request.write(data)

    def _close(self):
        """ Internal close """
        if not self.request.finished:
            # NYI: Debug only
            # logger.debug('%f: request.finish()', time.time())
            self.request.setResponseCode(200)
            self.request.finish()


class TwistedVapiResource(Resource):
    """ Twisted vapi resource """

    def __init__(self):
        """ Twisted vapi resource init """
        Resource.__init__(self)
        self.handler_map = {}

    def add_handler(self, content_type, protocol_handler):
        """
        add content handler

        :type  content_type: :class:`str`
        :param content_type: MIME content type
        :type  protocol_handler: :class:`vmware.vapi.protocol.server.transport.\
            async_protocol_handler.AsyncProtocolHandler`
        :param protocol_handler: protocol handler for this path
        """
        curr_handler = self.handler_map.get(content_type)
        if curr_handler:
            if curr_handler != protocol_handler:
                logger.error(
                    'Already registered. Failed to add handler for '
                    'content type %s', content_type)
                return False

        self.handler_map[content_type] = protocol_handler
        return True

    def render_POST(self, request):  # pylint: disable=invalid-name
        """
        Handle POST

        :type  request: :class:`twisted.web.http.Request`
        :param request: Twisted http request object
        """
        # Multiplex depends on content-type
        header_content_type = request.getHeader(b'Content-Type').decode('utf-8')
        tokens = header_content_type.split(';', 1)
        if len(tokens) > 1:
            content_type, _ = tokens
        else:
            content_type = tokens[0]
        handler = self.handler_map.get(content_type)
        if not handler:
            self.handle_error(b'', request, 500, b'Unsupported content')
            return

        request.setHeader(b'Server', b'Twisted/1.0')
        request.setHeader(b'Content-Type', content_type.encode('utf-8'))

        d = defer.Deferred()  # pylint: disable=invalid-name
        d.addCallback(self.handle_read)
        d.addCallback(self.handle_request, request, handler)
        d.addErrback(self.handle_error, request, 500, b'Request failed')
        # NYI: Add timeout for this connection
        # NYI: Add blacklist to protect from DoS attack

        d.callback(request)
        return twisted.web.server.NOT_DONE_YET

    def render_GET(self, request):  # pylint: disable=invalid-name
        """
        Handle HTTP GET

        :type  request: :class:`twisted.web.http.Request`
        :param request: Twisted http request object
        """
        header_content_type = request.getHeader(b'Content-Type').decode('utf-8')
        if header_content_type is None:
            content_type = ''
        else:
            tokens = header_content_type.split(';', 1)
            if len(tokens) > 1:
                content_type, _ = tokens
            else:
                content_type = tokens[0]
            request.setHeader(b'Content-Type', content_type.encode('utf-8'))

        request.setHeader(b'Server', b'Twisted/1.0')
        handler = self.handler_map.get(content_type)
        if not handler:
            self.handle_error(b'', request, 500, b'Unsupported content')
            return

        d = defer.Deferred()  # pylint: disable=invalid-name
        d.addCallback(self.handle_read)
        d.addCallback(self.handle_request, request, handler)
        d.addErrback(self.handle_error, request, 500, b'Request failed')
        # NYI: Add timeout for this connection
        # NYI: Add blacklist to protect from DoS attack

        d.callback(request)
        return twisted.web.server.NOT_DONE_YET

    def handle_error(self, failure, request, response_code, text_msg):  # pylint: disable=R0201
        """
        Handle error

        :type  failure: :class:`twisted.python.failure.Failure`
        :param failure: Twisted failure instance
        :type  request: :class:`twisted.web.http.Request`
        :param request: Twisted http request object
        :type  response_code: :class:`int`
        :param response_code: Http response code
        :type  text_msg: :class:`int`
        :param text_msg: Http error code
        """
        if not request.finished:
            logger.error(
                'handle_error: %s %d %s', failure, response_code, text_msg)
            request.setResponseCode(response_code)
            request.setHeader(b'Content-Type', b'text')
            request.write(text_msg)
            request.finish()

    def handle_read(self, request):  # pylint: disable=R0201
        """
        Handle read

        :type  request: :class:`twisted.web.http.Request`
        :param request: Twisted http request object
        """
        # logger.debug('handle_read:')
        # NYI: Chunking read, and send data one chunk at a time
        return request.content.read()

    def handle_request(self, request_msg, request, handler):  # pylint: disable=R0201
        """
        Handle request

        :type  request_msg: :class:`str`
        :param request_msg: Request msg
        :type  request: :class:`twisted.web.http.Request`
        :param request: Twisted http request object
        :type  handler: :class:`vmware.vapi.protocol.server.transport.\
            async_protocol_handler.AsyncProtocolHandler`
        :param handler: protocol handler for this addr
        """
        # logger.debug('handle_request: %s', request_msg)
        connection = TwistedHttpConnectionAdapter(request)
        data_handler = handler.get_data_handler(connection)
        data_handler.data_ready(request_msg)
        data_handler.data_end()


class TwistedHttpSite(Site):
    """ Twisted http site """
    HTTP_CONTENT_MAPPING = {'json': 'application/json',
                            'xml': 'text/xml',
                            '': ''}

    def __init__(self):
        """ Http site site init """
        self.root_resource = twisted.web.resource.Resource()
        twisted.web.server.Site.__init__(self, self.root_resource)

    def add_res_handler(self, path, msg_type, protocol_handler):
        """
        Add resource handler for a path

        :type  path: :class:`str`
        :param path: url path
        :type  msg_type: :class:`str`
        :param msg_type: transport msg type
        :type  protocol_handler: :class:`vmware.vapi.protocol.server.transport.\
            async_protocol_handler.AsyncProtocolHandler`
        :param protocol_handler: protocol handler for this path
        :rtype: :class:`bool`
        :return: True if added. False otherwise
        """
        content_type = self.HTTP_CONTENT_MAPPING.get(msg_type)
        if content_type is None:
            logger.error('Unsupported msg type: %s', msg_type)
            return False

        paths = [token for token in path.split('/') if token]
        if len(paths) == 0:
            logger.error('Cannot register root resource: %s', path)
            return False

        curr_resources = self.root_resource
        for node in paths[:-1]:
            node = node.encode('utf-8')
            node_entity = curr_resources.getStaticEntity(node)
            if node_entity:
                curr_resources = node_entity
            else:
                logger.debug('Adding child node %s', node)
                node_entity = Resource()
                curr_resources.putChild(node, node_entity)
                curr_resources = node_entity

        node = paths[-1].encode('utf-8')
        leave_entity = curr_resources.getStaticEntity(node)
        if isinstance(leave_entity, TwistedVapiResource):
            vapi_resource = leave_entity
        else:
            vapi_resource = TwistedVapiResource()
        if not vapi_resource.add_handler(content_type, protocol_handler):
            logger.error('Cannot re-register resource @ %s', path)
            return False
        logger.debug('Adding child node %s', node)
        curr_resources.putChild(node, vapi_resource)
        return True


class TwistedServer(ServerInterface):
    """ Twisted server """

    SUPPORTED_SCHEMES = ('vmware', 'vmwares', 'http', 'https')

    def __init__(self):
        """ Twisted server init """
        ServerInterface.__init__(self)
        self.lock = threading.Lock()
        self.handlers_map = {}

        # Protocol factories
        self.protocol_factories = {}

        self.main_thread_only = True

    def register_handler(self, addr, msg_type, protocol_handler, ssl_args=None):
        """
        Register protocol handler

        :type  addr: :class:`str`
        :param addr: addr url
        :type  msg_type: :class:`str`
        :param msg_type: protocol message type
        :type  protocol_handler: :class:`vmware.vapi.protocol.server.transport.\
            async_protocol_handler.AsyncProtocolHandler`
        :param protocol_handler: protocol handler for this addr
        :type  ssl_args: :class:`dict`
        :param ssl_args: ssl arguments
        """

        assert(protocol_handler)
        # addr is url
        scheme, host, port, _, _, path, _ = parse_addr_url(addr)
        # https / vmwares
        if scheme not in self.SUPPORTED_SCHEMES:
            logger.error('Unsupported url scheme: %s', addr)
            return

        # Get ssl context
        ssl_context = None
        if scheme == 'https' or scheme == 'vmwares':
            ssl_context = self.get_ssl_context(ssl_args)

        with self.lock:
            if scheme == 'http' or scheme == 'https':
                protocol_factory = self.protocol_factories.get((host, port))
                if not protocol_factory:
                    protocol_factory = TwistedHttpSite()
                    self.protocol_factories[(host, port)] = protocol_factory

                if not protocol_factory.add_res_handler(
                        path, msg_type, protocol_handler):
                    logger.error('Failed to add resource handler: %s', addr)
                    return
            elif scheme == 'vmware' or scheme == 'vmwares':
                protocol_factory = TwistedVmwareProtocolFactory(
                    protocol_handler)
            else:
                raise ValueError("Scheme %s is not supported" % scheme)

            self.handlers_map[(host, port)] = (protocol_factory, ssl_context)

    def get_ssl_context(self, ssl_args):
        """
        get ssl context

        :type  ssl_args: :class:`dict`
        :param ssl_args: ssl arguments
        :rtype: :class:`twisted.internet.ssl.DefaultOpenSSLContextFactory`
        :return: twisted ssl context factory instance
        """
        ssl_context = None
        ssl_ctx_args = {}
        if ssl_args:
            # Twisted don't support ca certs
            if ssl_args.get('keyfile', None):
                ssl_ctx_args['privateKeyFileName'] = ssl_args['keyfile']
            if ssl_args.get('certfile', None):
                ssl_ctx_args['certificateFileName'] = ssl_args['certfile']

        if len(ssl_ctx_args) > 0:
            logger.debug('ssl_ctx_args: %s', ssl_ctx_args)
            try:
                ssl_context = twisted.internet.ssl.DefaultOpenSSLContextFactory(
                    **ssl_ctx_args)
            except Exception:
                # Twisted without ssl support. Missing pyOpenSSL?
                logger.error('Twisted without ssl support. Missing pyOpenSSL?')
                raise
        else:
            logger.warning('Missing SSL context!')
        return ssl_context

    def serve_forever(self):
        """
        Server loop

        Note: Twisted limitation: Must be running from main thread
        """
        with self.lock:
            if len(self.handlers_map) == 0:
                logger.info('No handler registered. Server stopped')
                return

            for info, (protocol_factory, ssl_context) in six.iteritems(
                    self.handlers_map):
                host, port = info
                host = host if host else ''
                # isIPv6Address in twisted returns False if the addr has []
                # So, removing it from the addr.
                if host.startswith('[') and host.endswith(']'):
                    host = host[1:-1]
                logger.info('Listening on: %s %s', info, ssl_context)

                # interface='', listens on only IPv4 address
                # interface='::', listens on both IPv4 and IPv6 address
                # interface='<IPv4_addr>', listen on only IPv4 address
                # interface='<IPv6_addr>', listen on only IPv6 address
                if ssl_context:
                    twisted.internet.reactor.listenSSL(port,  # pylint: disable=E1101
                                                       protocol_factory,
                                                       ssl_context,
                                                       interface=host)
                else:
                    twisted.internet.reactor.listenTCP(port, protocol_factory,  # pylint: disable=E1101
                                                       interface=host)

        logger.info('twisted internet reactor started...')
        twisted.internet.reactor.run()  # pylint: disable=E1101
        logger.info('twisted internet reactor stopped')

    def shutdown(self):
        """ Server shutdown """

        try:
            twisted.internet.reactor.stop()  # pylint: disable=E1101
        except Exception:
            # Ignore all shutdown exceptions
            pass


def get_server():
    """
    Get twisted server

    :rtype: :class:`vmware.vapi.server.server_interface.ServerInterface`
    :return: subclass of ServerInterface
    """

    args = tuple()
    kwargs = {}

    # Twisted server MUST be singleton
    global _twisted_server  # pylint: disable=W0603
    with _server_lock:
        if not _twisted_server:
            _twisted_server = TwistedServer(*args, **kwargs)
        else:
            logger.warning(
                'Twisted server is singleton. This might not be what u wanted')
        server = _twisted_server
    return server



# -*- coding: utf-8 -*-
#---------------------------------------------------------------------------
# Copyright 2023 VMware, Inc.  All rights reserved.

# AUTO GENERATED FILE -- DO NOT MODIFY!
#
# vAPI stub file for package com.vmware.vcenter.deployment.install.
#---------------------------------------------------------------------------

"""


"""

__author__ = 'VMware, Inc.'
__docformat__ = 'restructuredtext en'

import sys
from warnings import warn

from vmware.vapi.bindings import type
from vmware.vapi.bindings.converter import TypeConverter
from vmware.vapi.bindings.enum import Enum
from vmware.vapi.bindings.error import VapiError
from vmware.vapi.bindings.struct import VapiStruct
from vmware.vapi.bindings.stub import (
    ApiInterfaceStub, StubFactoryBase, VapiInterface)
from vmware.vapi.bindings.common import raise_core_exception
from vmware.vapi.data.validator import (UnionValidator, HasFieldsOfValidator)
from vmware.vapi.exception import CoreException
from vmware.vapi.lib.constants import TaskType
from vmware.vapi.lib.rest import OperationRestMetadata


class RemotePsc(VapiInterface):
    """
    The ``RemotePsc`` class provides methods to check if the deployed vCenter
    Server can register with the remote PSC. This class was added in vSphere
    API 6.7.
    """

    _VAPI_SERVICE_ID = 'com.vmware.vcenter.deployment.install.remote_psc'
    """
    Identifier of the service in canonical form.
    """
    def __init__(self, config):
        """
        :type  config: :class:`vmware.vapi.bindings.stub.StubConfiguration`
        :param config: Configuration to be used for creating the stub.
        """
        VapiInterface.__init__(self, config, _RemotePscStub)
        self._VAPI_OPERATION_IDS = {}


    def check(self,
              spec,
              ):
        """
        Checks whether the remote PSC is reachable and the deployed vCenter
        Server can be registered with the remote PSC. This method was added in
        vSphere API 6.7.

        :type  spec: :class:`com.vmware.vcenter.deployment_client.RemotePscSpec`
        :param spec: Information to connect to the remote PSC.
        :rtype: :class:`com.vmware.vcenter.deployment_client.CheckInfo`
        :return: Information about the success or failure of the checks that were
            performed.
        :raise: :class:`com.vmware.vapi.std.errors_client.Unauthenticated` 
            if the caller is not authenticated.
        :raise: :class:`com.vmware.vapi.std.errors_client.Unauthenticated` 
            if external PSC credentials are not valid when configuring a
            VCSA_EXTERNAL appliance.
        :raise: :class:`com.vmware.vapi.std.errors_client.InvalidArgument` 
            if passed arguments are invalid.
        :raise: :class:`com.vmware.vapi.std.errors_client.NotAllowedInCurrentState` 
            if the appliance is not in INITIALIZED state.
        """
        return self._invoke('check',
                            {
                            'spec': spec,
                            })
class _RemotePscStub(ApiInterfaceStub):
    def __init__(self, config):
        # properties for check operation
        check_input_type = type.StructType('operation-input', {
            'spec': type.ReferenceType('com.vmware.vcenter.deployment_client', 'RemotePscSpec'),
        })
        check_error_dict = {
            'com.vmware.vapi.std.errors.unauthenticated':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'Unauthenticated'),
            'com.vmware.vapi.std.errors.invalid_argument':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'InvalidArgument'),
            'com.vmware.vapi.std.errors.not_allowed_in_current_state':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'NotAllowedInCurrentState'),

        }
        check_input_value_validator_list = [
        ]
        check_output_validator_list = [
        ]
        check_rest_metadata = OperationRestMetadata(
            http_method='POST',
            url_template='/vcenter/deployment/install/remote-psc?action=check',
            path_variables={
            },
             header_parameters={
               },
            query_parameters={
            }
        )

        operations = {
            'check': {
                'input_type': check_input_type,
                'output_type': type.ReferenceType('com.vmware.vcenter.deployment_client', 'CheckInfo'),
                'errors': check_error_dict,
                'input_value_validator_list': check_input_value_validator_list,
                'output_validator_list': check_output_validator_list,
                'task_type': TaskType.NONE,
            },
        }
        rest_metadata = {
            'check': check_rest_metadata,
        }
        ApiInterfaceStub.__init__(
            self, iface_name='com.vmware.vcenter.deployment.install.remote_psc',
            config=config, operations=operations, rest_metadata=rest_metadata,
            is_vapi_rest=True)


class StubFactory(StubFactoryBase):
    _attrs = {
        'RemotePsc': RemotePsc,
        'psc': 'com.vmware.vcenter.deployment.install.psc_client.StubFactory',
    }




"""
REST Handler for WSGI application
"""

__author__ = 'VMware, Inc.'
__copyright__ = 'Copyright 2015-2020, 2023 VMware, Inc.  All rights reserved. -- VMware Confidential'  # pylint: disable=line-too-long

import base64
import collections
import decimal
import json
import six
import werkzeug

from com.vmware.vapi.metadata.metamodel_client import (
    Component, Structure, Type, Enumeration,
    GenericInstantiation, StructureInfo)
from vmware.vapi.bindings.task_helper import get_task_operation_name
from vmware.vapi.common.context import set_context, clear_context
from vmware.vapi.core import ApplicationContext, ExecutionContext
from vmware.vapi.data.serializers.cleanjson import DataValueConverter
from vmware.vapi.data.type import Type as DataType
from vmware.vapi.data.value import (
    data_value_factory, ListValue, StructValue,
    StringValue, OptionalValue, VoidValue, SecretValue)
from vmware.vapi.lib.constants import (
    HTTP_ACCEPT_LANGUAGE, HTTP_FORMAT_LOCALE, HTTP_TIMEZONE, LOCALE,
    MAP_ENTRY, OPERATION_INPUT, OPID, RestAnnotations,
    RestAnnotationType, REST_OP_ID_HEADER, TASK_REST_QUERY_PARAM)
from vmware.vapi.lib.context import create_default_application_context
from vmware.vapi.lib.load import dynamic_import_list
from vmware.vapi.lib.log import get_vapi_logger
from vmware.vapi.protocol.client.local_connector import get_local_connector
from vmware.vapi.security.session import (
    REST_SESSION_ID_KEY, REQUIRE_HEADER_AUTHN)
from vmware.vapi.stdlib.client.factories import StubConfigurationFactory
from vmware.vapi.exception import CoreException

from .lib import vapi_to_http_error_map
from .rules import RoutingRuleGenerator

logger = get_vapi_logger(__name__)


class MetadataStore(object):
    """
    Helper class to process the metamodel metadata and
    provide a convenient way to access the data.
    """

    class OperationSummary(object):
        """
        Helper class to contain only useful metamodel metadata of an operation
        """
        def __init__(self, param_info_map, path_variables_map=None,
                     request_mapping_metadata=None, verb_metadata=None,
                     query_variables_map=None, header_variables_map=None,
                     success_reponse_code=None, response_headers_map=None,
                     response_body_name=None):
            """
            :type  param_info_map:
                :class:`collections.OrderedDict` of :class:`str` and
                :class:`com.vmware.vapi.metadata.metamodel_client.FieldInfo`
            :param param_info_map: Map of parameter name to its metamodel
                metadata
            :type  path_variables_map: :class:`dict` of :class:`str` and
                :class:`str`
            :param path_variables_map: Map of path variable name to canonical
                name of the parameters that have the PathVariable annotation
            :type  request_mapping_metadata:
                :class:`com.vmware.vapi.metadata.metamodel_provider.ElementMap`
            :param request_mapping_metadata: Metamodel metadata of
                RequestMapping annotation on the operation
            :type  verb_metadata:
                :class:`com.vmware.vapi.metadata.metamodel_provider.ElementMap`
            :param verb_metadata: Metamodel metadata of
                Verb annotation on the operation
            :type  query_variables_map:
                :class:`com.vmware.vapi.metadata.metamodel_provider.ElementMap`
            :param query_variables_map: Metamodel metadata of
                dispatch query variable in Verb annotation on the operation
            :type  header_variables_map:
                :class:`com.vmware.vapi.metadata.metamodel_provider.ElementMap`
            :param header_variables_map: Metamodel metadata of
                dispatch header variable in Verb annotation on the operation
            :type  success_reponse_code:
                :class:`int`
            :param success_response_code: Response code for success
            :type  response_headers_map:
                :class:`dict` of :class:`str` and :class:`str`
            :param response_headers_map: Map of response field name to
                http header name
            :type  response_body_name:
                :class:`str`
            :param response_body_name: Response field name has
                Body annotation, will only return the value of the field
            """
            self.param_info_map = param_info_map
            self.path_variables_map = path_variables_map
            self.request_mapping_metadata = request_mapping_metadata
            self.verb_metadata = verb_metadata
            self.query_variables_map = query_variables_map
            self.header_variables_map = header_variables_map
            self.success_response_code = success_reponse_code
            self.response_headers_map = response_headers_map
            self.response_body_name = response_body_name

        def has_request_mapping_metadata(self):
            """
            Tells whether this operation has RequestMapping annotation
            """
            return self.request_mapping_metadata is not None

        def has_verb_metadata(self):
            """
            Tells whether this operation has Verb annotation
            """
            return self.verb_metadata is not None

    def __init__(self, api_provider):
        """
        Initialize MetadataStore

        :type  api_provider: :class:`vmware.vapi.core.ApiProvider`
        :param api_provider: ApiProvider to get the metadata
        """
        self.structure_map = {}
        self.enumeration_map = {}
        self.service_map = {}
        # Map of {struct_id: {error_id: {code: http_status, reason: ''}}}
        self.all_errors_response_code_map = {}
        # Map of {struct_id: {field_name: header_name}}
        self.all_response_header_map = {}
        # Map of {struct_id: field_name}
        self.all_response_body_map = {}
        self._build(api_provider)

    def _build(self, api_provider):
        """
        Get the metamodel metadata and process it

        :type  api_provider: :class:`vmware.vapi.core.ApiProvider`
        :param api_provider: ApiProvider to get the metadata
        """
        local_connector = get_local_connector(api_provider)
        stub_config = StubConfigurationFactory.new_std_configuration(
            local_connector)
        component_svc = Component(stub_config)
        components = component_svc.list()
        for component_id in components:
            component_data = component_svc.get(component_id)
            self._create_structure_map(component_data.info)
            self._create_operation_map(component_data.info)

    def _create_structure_map(self, component_info):
        """
        Process the metamodel component information. Scan the packages
        and services for structures and enumerations and store into dictionaries
        for fast lookup.

        :type  component_info:
               :class:`com.vmware.vapi.metadata.metamodel_client.ComponentInfo`
        :param component_info: Metamodel component information to be processed
        """
        for package_info in six.itervalues(component_info.packages):
            for structure_id, structure_info in six.iteritems(
                                                package_info.structures):
                self._process_structure_info(structure_id, structure_info)
            for enumeration_id, enumeration_info in six.iteritems(
                                                    package_info.enumerations):
                self.enumeration_map[enumeration_id] = enumeration_info
            for service_id, service_info in six.iteritems(
                                            package_info.services):
                self._process_service_info(service_id, service_info)


    def _create_operation_map(self, component_info):
        """
        Process the metamodel component information. Scan services for
        operations and store into dictionaries for fast lookup.

        :type  component_info:
               :class:`com.vmware.vapi.metadata.metamodel_client.ComponentInfo`
        :param component_info: Metamodel component information to be processed
        """
        for package_info in six.itervalues(component_info.packages):
            for service_id , service_info in six.iteritems(package_info.services):
                self.service_map[service_id] = {
                    operation_id: self._create_operation_summary(operation_info,
                                                                 service_id)
                        for operation_id, operation_info in six.iteritems(
                                                    service_info.operations)
                }

    def _get_header_and_query_variables(self, param_info_map):
        """
        Get the annotation @Header or @Query parameter alias from metadata.

        :type  param_info_map:
            :class:`collections.OrderedDict` of :class:`str` and
            :class:`com.vmware.vapi.metadata.metamodel_client.FieldInfo`
        :param param_info_map: Map of parameter name to its metamodel metadata
        :rtype: :class: `tuple`
        :return: Tuple containing the @Header and @Query annotation parameter
            alias.
        """
        header_variables = {}
        query_variables = {}
        for param_name, param_info in six.iteritems(param_info_map):
            resource_id = None
            # Get the structure body resource_id from param_info.
            if param_info.type.category == Type.Category.USER_DEFINED:
                resource_id = param_info.type.user_defined_type.resource_id
            # Get the optional structure resource_id from param_info.
            elif param_info.type.category == Type.Category.GENERIC:
                generic_type = param_info.type.generic_instantiation.generic_type  # pylint: disable=line-too-long
                if generic_type == GenericInstantiation.GenericType.OPTIONAL:
                    element_type = param_info.type.generic_instantiation.element_type  # pylint: disable=line-too-long
                    if element_type.category == Type.Category.USER_DEFINED:
                        resource_id = element_type.user_defined_type.resource_id

            # resource_id==None indicates that the parameter is not a struct
            # type and it might be an enum type
            if resource_id is None or resource_id in self.enumeration_map:
                # Get the @Header annotation parameter alias from metadata.
                if RestAnnotations.HEADER_ELEMENT in param_info.metadata:
                    alias = param_info.metadata[
                        RestAnnotations.HEADER_ELEMENT].elements[
                            RestAnnotations.NAME_ELEMENT].string_value
                    header_variables[
                        alias.title().replace('_', '-')] = param_name
                # Get the @Query annotation parameter alias from metadata.
                if RestAnnotations.QUERY_VARIABLE in param_info.metadata:
                    alias = param_info.metadata[
                        RestAnnotations.QUERY_VARIABLE].elements[
                            RestAnnotations.NAME_ELEMENT].string_value
                    query_variables[alias] = param_name
            # resource_id!=None or not in enumerations, indicates that the parameter is of struct type
            else:
                for name, info in six.iteritems(
                                            self.structure_map[resource_id]):
                    if info.type.category != Type.Category.USER_DEFINED:
                        # Get the @Header annotation parameter alias from
                        # the metadata of the structure
                        if RestAnnotations.HEADER_ELEMENT in info.metadata:
                            alias = info.metadata[
                                RestAnnotations.HEADER_ELEMENT].elements[
                                    RestAnnotations.NAME_ELEMENT].string_value
                            name = '%s.%s' % (param_name, name)
                            header_variables[
                                alias.title().replace('_', '-')] = name
                        # Get the @Query annotation parameter alias from
                        # the metadata of the structure
                        if RestAnnotations.QUERY_VARIABLE in info.metadata:
                            alias = info.metadata[
                                RestAnnotations.QUERY_VARIABLE].elements[
                                    RestAnnotations.NAME_ELEMENT].string_value
                            name = '%s.%s' % (param_name, name)
                            query_variables[
                                '%s.%s' % (param_name, alias)] = name

        return header_variables, query_variables

    def _translate_param_name(self, name):
        """
        Changes '__' to '-' in the parameter name. Before registering the
        metadata dashes are changed because werkzeug doesn't allow dashes.

        :type  name: :class:`str`
        :param component_info: parameter name
        :rtype: :class:`str`
        :return: parameter name with dashes instead of '__'
                 If there is nothing to change the provided name is returned
        """
        return name.replace('-', '__')

    def _create_operation_summary(self, operation_info, service_id):
        """
        Generate a summary of the metamodel operation information.

        :type  component_info:
               :class:`com.vmware.vapi.metadata.metamodel_client.OperationInfo`
        :param component_info: Metamodel metadata of an operation
        :rtype: :class:`vmware.vapi.server.rest_handler.MetadataStore.\
            OperationSummary`
        :return: Class containing parameters metadata and annotations in dicts
        """
        param_list = [(param_info.name, param_info)
                      for param_info in operation_info.params]
        param_info_map = collections.OrderedDict(param_list)
        path_variables_map = None
        query_variables_map = None
        header_variables_map = None
        success_response_code = None
        response_headers_map = None
        response_body_name = None
        verb_info = None

        if RestAnnotations.REQUEST_MAPPING in operation_info.metadata.keys():
            # Create a map of value in PathVariable annotation to
            # canonical name of the parameter
            path_variables_map = {
                self._translate_param_name(param_info.metadata[
                    RestAnnotations.PATH_VARIABLE].elements[
                    RestAnnotations.VALUE_ELEMENT].string_value): param_name
                for param_name, param_info in six.iteritems(param_info_map)
                if RestAnnotations.PATH_VARIABLE in param_info.metadata
            }
        elif service_id != 'com.vmware.cis.session':
            if RestAnnotations.VERB_GET in operation_info.metadata.keys():
                verb_info = {RestAnnotations.VERB_GET:
                             operation_info.metadata.get(RestAnnotations.VERB_GET)}  # pylint: disable=line-too-long
            elif RestAnnotations.VERB_POST in operation_info.metadata.keys():
                verb_info = {RestAnnotations.VERB_POST:
                             operation_info.metadata.get(RestAnnotations.VERB_POST)}  # pylint: disable=line-too-long
            elif RestAnnotations.VERB_PUT in operation_info.metadata.keys():
                verb_info = {RestAnnotations.VERB_PUT:
                             operation_info.metadata.get(RestAnnotations.VERB_PUT)}  # pylint: disable=line-too-long
            elif RestAnnotations.VERB_PATCH in operation_info.metadata.keys():
                verb_info = {RestAnnotations.VERB_PATCH:
                             operation_info.metadata.get(
                                 RestAnnotations.VERB_PATCH)}
            elif RestAnnotations.VERB_DELETE in operation_info.metadata.keys():
                verb_info = {RestAnnotations.VERB_DELETE:
                             operation_info.metadata.get(
                                RestAnnotations.VERB_DELETE)}

        if verb_info is not None:
            header_variables_map, query_variables_map = \
                self._get_header_and_query_variables(param_info_map)

            if RestAnnotations.RESPONSE_ELEMENT in operation_info.metadata.keys():  # pylint: disable=line-too-long
                success_response_code = int(operation_info.metadata.get(
                        RestAnnotations.RESPONSE_ELEMENT).elements['code'].string_value)  # pylint: disable=line-too-long

            response_headers_map = {}
            result_type = operation_info.output.type
            # response header can only be annotated within a structure
            # and the structure belongs to user defined category type
            if result_type.category == Type.Category.USER_DEFINED and \
               result_type.user_defined_type.resource_type == \
               Structure.RESOURCE_TYPE:
                structure_name = \
                    result_type.user_defined_type.resource_id
                if structure_name in self.all_response_header_map:
                    header_map = \
                        self.all_response_header_map.get(structure_name)
                    response_headers_map.update(header_map)
                if structure_name in self.all_response_body_map:
                    response_body_name = \
                        self.all_response_body_map.get(structure_name)

        # XXX Instead of long if else above the code should change to line
        # below if metamodel metadata changes to single key
        #verb_info = operation_info.metadata.get(RestAnnotations.REST_DISPATCH)
        return self.OperationSummary(
            param_info_map,
            path_variables_map,
            operation_info.metadata.get(RestAnnotations.REQUEST_MAPPING),
            verb_info,
            query_variables_map,
            header_variables_map,
            success_response_code,
            response_headers_map,
            response_body_name
        )

    def _process_service_info(self, service_id, service_info):
        """
        Process the metamodel service information. Scan the services
        for operations, structures and enumerations and store into dictionaries
        for fast lookup.

        :type  service_id: :class:`str`
        :param service_id: Identifier of the service.
        :type  service_info:
               :class:`com.vmware.vapi.metadata.metamodel_client.ServiceInfo`
        :param service_info: Metamodel service information to be processed
        """
        for structure_id, structure_info in six.iteritems(
                                            service_info.structures):
            self._process_structure_info(structure_id, structure_info)

        for enumeration_id, enumeration_info in six.iteritems(
                                                service_info.enumerations):
            self.enumeration_map[enumeration_id] = enumeration_info

    def _process_structure_info(self, structure_id, structure_info):
        """
        Process the metamodel structure information. Scan the structures for
        for fields and enumerations and store into dictionaries
        for fast lookup.

        :type  structure_id: :class:`str`
        :param structure_id: Identifier of the structure.
        :type  structure_info:
               :class:`com.vmware.vapi.metadata.metamodel_client.StructureInfo`
        :param structure_info: Metamodel structure information to be processed
        """
        field_map = {}
        field_header_map = {}
        for field_info in structure_info.fields:
            field_map[field_info.name] = field_info
            if RestAnnotations.HEADER_ELEMENT in field_info.metadata.keys():
                header_name = field_info.metadata.get(
                    RestAnnotations.HEADER_ELEMENT).elements['name'].string_value  # pylint: disable=line-too-long
                field_header_map[field_info.name] = header_name
            if RestAnnotations.BODY_ELEMENT in field_info.metadata.keys():
                self.all_response_body_map[structure_id] = field_info.name

        self.structure_map[structure_id] = field_map

        if field_header_map:
            self.all_response_header_map[structure_id] = field_header_map

        for enumeration_id, enumeration_info in six.iteritems(
                                                structure_info.enumerations):
            self.enumeration_map[enumeration_id] = enumeration_info

        if structure_info.type == StructureInfo.Type.ERROR \
                and RestAnnotations.RESPONSE_ELEMENT in structure_info.metadata.keys():  # pylint: disable=line-too-long
            error_element = structure_info.metadata.get(RestAnnotations.RESPONSE_ELEMENT)  # pylint: disable=line-too-long
            self.all_errors_response_code_map[structure_id] = \
                int(error_element.elements['code'].string_value)


class URLValueDeserializer(object):
    """
    Deserialize the parameters provided in the HTTP query string
    into a dictionary.

    For example:
        /rest/vmodl/test/uber/rest/filter?
            list_string=string1&
            list_string=string2&
            list_string=string3&
            struct.string_field=string&
            struct.binary_field=aGVsbG8=&
            struct.boolean_field=True&
            struct.long_field=10&
            struct.struct_field.string_field=string&
            struct.struct_field.long_field=10&
            struct.optional_string=string&
            struct.list_string.1=string1&
            struct.list_string.2=string2&
            struct.list_list_long.1.1=11&
            struct.list_list_long.1.2=12&
            struct.list_list_long.2.1=21&
            struct.list_list_long.2.2=22&
            struct.map_simple.1.key=stringkey&
            struct.map_simple.1.value=stringvalue&
            struct.map_struct.1.key=stringkey&
            struct.map_struct.1.value.string_field=string&
            struct.map_struct.1.value.long_field=10

    1. Top level members of the request complex type are referenced by their
       name.
    2. Referencing nested complex structures will use "." and concatenation of
       names.
    3. Whenever arrays of given type are required we will use param.n notation
       to identify the instance number n of the object.

    Deserialized version of this query string will be:
    {
        "list_string": [
            "string1",
            "string2",
            "string3"
        ],
        "struct": {
            "string_field": "string",
            "binary_field": "aGVsbG8=",
            "boolean_field": True,
            "long_field": 10,
            "struct_field": {
                "long_field": 10,
                "string_field": "string"
            },
            "optional_string": "string",
            "list_string": [
                "string1",
                "string2"
            ],
            "list_list_long": [
                [
                    11,
                    12
                ],
                [
                    21,
                    22
                ]
            ],
            "map_simple": [
                {
                    "key": "stringkey",
                    "value": "stringvalue"
                }
            ],
            "map_struct": [
                {
                    "key": "stringkey",
                    "value": {
                        "long_field": 10,
                        "string_field": "string"
                    }
                }
            ]
        }
    }
    """

    @staticmethod
    def deserialize(query_string, mapping_type=RestAnnotationType.VERB):
        """
        Deserialize the given query string into a python dictionary.

        :type  query_string: :class:`str`
        :param query_string: HTTP query string containing parameters.
        :type  mapping_type:
            :class:`vmware.vapi.lib.constants.RestAnnotationType`
        :param mapping_type: Rest annotation type needed for handling map input
        :rtype: :class:`dict` of :class:`str` and :class:`object`
        :return: Python dictionary deserialized from query string.
        """
        # Convert query_string to a key-value pair dictionary
        key_value_pairs = collections.OrderedDict()
        query_string = query_string.decode()
        for item in query_string.split('&'):
            if '=' not in item:
                continue

            pos = item.index('=')
            key = item[:pos]
            value = item[pos + 1:]
            if value != '':
                if key not in key_value_pairs:
                    key_value_pairs[key] = value
                elif isinstance(key_value_pairs[key], six.text_type):
                    key_value_pairs[key] = [key_value_pairs[key], value]
                elif isinstance(key_value_pairs[key], list):
                    key_value_pairs[key].append(value)

        # Deserialize query_string
        query_string_dict = {}
        for k, v in six.iteritems(key_value_pairs):
            tokens = k.split('.')

            # The core idea here is to split the tokens in
            # a.b.c.d type of string and go from left to right.
            # At each step, the next token will be either put
            # at a key in the dictionary or element in a list if
            # the token is a digit.

            # So, at each token, we should look ahead the next token,
            # and if the next token is a digit, then create an empty
            # list or return an existing list where it should be added
            # as element. If the next token is a non-digit, then create
            # a new dictionary or return an existing dictionary where
            # the token should be added as a key.

            current_value = query_string_dict
            key = None
            for i, token in enumerate(tokens):

                # Lookahead the next token
                next_token = None
                if i + 1 < len(tokens):
                    next_token = tokens[i + 1]

                if key is None and mapping_type == RestAnnotationType.VERB:
                    if next_token is None:
                        key = '.'.join(tokens)
                        current_value.setdefault(key, v)
                        current_value = current_value[key]
                        break
                    elif next_token.isdigit():
                        key = '.'.join(tokens[:i + 1])
                        current_value.setdefault(key, [])
                        current_value = current_value[key]
                else:
                    # Next token is the last token
                    if next_token is None:
                        if isinstance(current_value, list):
                            current_value.append(v)  # pylint: disable=E1101
                        elif isinstance(current_value, dict):
                            current_value[token] = v
                    # Next token is not the last token
                    else:
                        # If next token is a digit, create array as placeholder
                        # otherwise a dictionary
                        next_token_type = [] if next_token.isdigit() else {}

                        if isinstance(current_value, list):
                            index = int(token)
                            if index > len(current_value) + 1:
                                msg = (
                                    'Element with index %d is expected, but got'
                                    ' %d' % (len(current_value) + 1, index))
                                logger.error(msg)
                                raise werkzeug.exceptions.BadRequest(msg)
                            elif index == len(current_value) + 1:
                                # pylint: disable=E1101
                                current_value.append(next_token_type)
                            next_value = current_value[index - 1]
                        elif isinstance(current_value, dict):
                            next_value = current_value.setdefault(
                                token, next_token_type)
                        current_value = next_value

        return query_string_dict


class DataValueDeserializer(object):
    """
    Convert from Python dictionary deserialized from a JSON object
    (or) from HTTP URL query string to DataValue.
    """
    _builtin_type_map = {
        Type.BuiltinType.VOID: DataType.VOID,
        Type.BuiltinType.BOOLEAN: DataType.BOOLEAN,
        Type.BuiltinType.LONG: DataType.INTEGER,
        Type.BuiltinType.DOUBLE: DataType.DOUBLE,
        Type.BuiltinType.STRING: DataType.STRING,
        Type.BuiltinType.BINARY: DataType.BLOB,
        Type.BuiltinType.SECRET: DataType.SECRET,
        Type.BuiltinType.DATE_TIME: DataType.STRING,
        Type.BuiltinType.ID: DataType.STRING,
        Type.BuiltinType.URI: DataType.STRING,
        Type.BuiltinType.ANY_ERROR: DataType.ANY_ERROR,
    }
    _builtin_native_type_map = {
        Type.BuiltinType.BOOLEAN: bool,
        Type.BuiltinType.LONG: int,
        Type.BuiltinType.DOUBLE: decimal.Decimal,
        Type.BuiltinType.STRING: six.text_type,
        Type.BuiltinType.BINARY: six.text_type,
        Type.BuiltinType.SECRET: six.text_type,
        Type.BuiltinType.DATE_TIME: six.text_type,
        Type.BuiltinType.ID: six.text_type,
        Type.BuiltinType.URI: six.text_type,
    }

    def __init__(self, metadata):
        """
        Initialize DataValueDeserializer

        :type  metadata: :class:`vmware.vapi.server.rest_handler.MetadataStore`
        :param metadata: Object that contains the metamodel metadata of
            all the services.
        """
        self._metadata = metadata
        self._category_map = {
            Type.Category.BUILTIN: self.visit_builtin,
            Type.Category.USER_DEFINED: self.visit_user_defined,
            Type.Category.GENERIC: self.visit_generic
        }
        self._generic_map = {
            GenericInstantiation.GenericType.OPTIONAL: self.visit_optional,
            GenericInstantiation.GenericType.LIST: self.visit_list,
            GenericInstantiation.GenericType.SET: self.visit_list,
            GenericInstantiation.GenericType.MAP: self.visit_map,
        }

    def get_metadata(self):
        """
        Return the metamodel metadata

        :rtype: :class:`vmware.vapi.server.rest_handler.MetadataStore`
        :return: Metamodel metadata
        """
        return self._metadata

    def visit_builtin(self, type_info, json_value):
        """
        Deserialize a primitive value

        :type  type_info:
            :class:`com.vmware.vapi.metadata.metamodel_client.Type`
        :param type_info: Metamodel type information
        :type  json_value: :class:`object`
        :param json_value: Value to be visited
        :rtype: :class:`vmware.vapi.data.value.DataValue`
        :return: DataValue created using the input
        """
        try:
            builtin_type = type_info.builtin_type
            native_type = self._builtin_native_type_map.get(
                                  builtin_type)
            if native_type == six.text_type:
                if type_info.builtin_type == Type.BuiltinType.BINARY:
                    # For Binary types, we need to convert unicode to bytes
                    base64_encoded_value = json_value.encode()
                    native_value = base64.b64decode(base64_encoded_value)
                else:
                    native_value = json_value
            elif native_type == int:
                native_value = int(json_value)
            elif native_type == decimal.Decimal:
                native_value = decimal.Decimal(json_value)
            elif native_type == bool:
                if isinstance(json_value, bool):
                    native_value = json_value
                elif json_value.lower() == 'true':
                    native_value = True
                elif json_value.lower() == 'false':
                    native_value = False
                else:
                    msg = 'Expected boolean value, but got %s' % json_value
                    logger.error(msg)
                    raise werkzeug.exceptions.BadRequest(msg)
            elif native_type is None:
                if builtin_type == Type.BuiltinType.DYNAMIC_STRUCTURE:
                    return DataValueConverter.convert_to_data_value(
                                                    json.dumps(json_value))
                elif builtin_type == Type.BuiltinType.OPAQUE:
                    if isinstance(json_value, six.text_type):
                        builtin_type = Type.BuiltinType.STRING
                        native_value = json_value
                    elif isinstance(json_value, bool):
                        builtin_type = Type.BuiltinType.BOOLEAN
                        native_value = json_value
                    elif isinstance(json_value, int):
                        builtin_type = Type.BuiltinType.LONG
                        native_value = json_value
                    elif isinstance(json_value, decimal.Decimal):
                        builtin_type = Type.BuiltinType.DOUBLE
                        native_value = decimal.Decimal(json_value)
                    else:
                        msg = ('Expected string or boolean or long or double '
                               'value, but got %s' % json_value)
                        logger.error(msg)
                        raise werkzeug.exceptions.BadRequest(msg)

            data_type = self._builtin_type_map[builtin_type]
            return data_value_factory(data_type, native_value)
        except KeyError:
            msg = ('Could not process the request, '
                   'builtin type %s is not supported' % type_info.builtin_type)
            logger.exception(msg)
            raise werkzeug.exceptions.InternalServerError(msg)

    def visit_user_defined(self, type_info, json_value):
        """
        Deserialize a user defined value

        :type  type_info:
            :class:`com.vmware.vapi.metadata.metamodel_client.Type`
        :param type_info: Metamodel type information
        :type  json_value: :class:`object`
        :param json_value: Value to be visited
        :rtype: :class:`vmware.vapi.data.value.StructValue` or
            :class:`vmware.vapi.data.value.StringValue`
        :return: DataValue created using the input
        """
        user_defined_type_info = type_info.user_defined_type
        if user_defined_type_info.resource_type == Enumeration.RESOURCE_TYPE:
            return StringValue(json_value)
        elif user_defined_type_info.resource_type == Structure.RESOURCE_TYPE:
            structure_info = self._metadata.structure_map[
                user_defined_type_info.resource_id]
            struct_value = StructValue(
                name=user_defined_type_info.resource_id)
            for field_name, field_value in six.iteritems(json_value):
                try:
                    field_info = structure_info[str(field_name)]
                    field_data_value = self.visit(field_info.type, field_value)
                    struct_value.set_field(field_name, field_data_value)
                except KeyError:
                    msg = 'Unexpected field \'%s\' in request' % field_name
                    logger.error(msg)
                    raise werkzeug.exceptions.BadRequest(msg)
                except CoreException as e:
                    msg = 'Unable to parse property with name \'%s\': %s' % (field_name, e)
                    logger.exception(msg)
                    raise werkzeug.exceptions.BadRequest(msg)
            return struct_value
        else:
            msg = ('Could not process the request,'
                   'user defined type %s is not supported' %
                   user_defined_type_info.resource_type)
            logger.error(msg)
            raise werkzeug.exceptions.InternalServerError(msg)

    def visit_optional(self, type_info, json_value):
        """
        Deserialize an optional value

        :type  type_info:
            :class:`com.vmware.vapi.metadata.metamodel_client.Type`
        :param type_info: Metamodel type information
        :type  json_value: :class:`object`
        :param json_value: Value to be visited
        :rtype: :class:`vmware.vapi.data.value.OptionalValue`
        :return: DataValue created using the input
        """
        if json_value is not None:
            element_value = self.visit(type_info.element_type, json_value)
            return OptionalValue(element_value)
        else:
            return OptionalValue()

    def visit_list(self, type_info, json_value):
        """
        Deserialize a list value

        :type  type_info:
            :class:`com.vmware.vapi.metadata.metamodel_client.Type`
        :param type_info: Metamodel type information
        :type  json_value: :class:`object`
        :param json_value: Value to be visited
        :rtype: :class:`vmware.vapi.data.value.ListValue`
        :return: DataValue created using the input
        """
        if not isinstance(json_value, list):
            msg = 'Excepted list, but got %s' % type(json_value).__name__
            logger.error(msg)
            raise werkzeug.exceptions.BadRequest(msg)
        return ListValue([
            self.visit(type_info.element_type, value) for value in json_value])

    def visit_map(self, type_info, json_value):
        """
        Deserialize a map value

        :type  type_info:
            :class:`com.vmware.vapi.metadata.metamodel_client.Type`
        :param type_info: Metamodel type information
        :type  json_value: :class:`object`
        :param json_value: Value to be visited
        :rtype: :class:`vmware.vapi.data.value.StructValue`
        :return: DataValue created using the input
        """
        # For new Rest annotations map deserializes to StructValue instead of
        # ListValue for old annotations
        if (self.mapping_type == RestAnnotationType.VERB):
            if not isinstance(json_value, dict):
                msg = 'Excepted dict, but got %s' % type(json_value).__name__
                logger.error(msg)
                raise werkzeug.exceptions.BadRequest(msg)

            struct_value = StructValue()
            for field_name, field_value in six.iteritems(json_value):
                try:
                    field_data_value = self.visit(type_info.map_value_type,
                                                  field_value)
                    struct_value.set_field(field_name, field_data_value)
                except KeyError:
                    msg = 'Unexpected field \'%s\' in request' % field_name
                    logger.error(msg)
                    raise werkzeug.exceptions.BadRequest(msg)
            return struct_value
        else:
            if not isinstance(json_value, list):
                msg = 'Excepted list, but got %s' % type(json_value).__name__
                logger.error(msg)
                raise werkzeug.exceptions.BadRequest(msg)
            try:
                return ListValue(
                    [StructValue(
                        name=MAP_ENTRY,
                        values={
                            'key': self.visit(
                                type_info.map_key_type, value['key']),
                            'value': self.visit(
                                type_info.map_value_type, value['value'])
                        })
                        for value in json_value])
            except KeyError as e:
                msg = 'Invalid Map input, missing %s' % e
                logger.error(msg)
                raise werkzeug.exceptions.BadRequest(msg)

    def visit_generic(self, type_info, json_value):
        """
        Deserialize a list/optional/map value

        :type  type_info:
            :class:`com.vmware.vapi.metadata.metamodel_client.Type`
        :param type_info: Metamodel type information
        :type  json_value: :class:`object`
        :param json_value: Value to be visited
        :rtype: :class:`vmware.vapi.data.value.OptionalValue` or
            :class:`vmware.vapi.data.value.ListValue` or
            :class:`vmware.vapi.data.value.StructValue`
        :return: DataValue created using the input
        """
        try:
            generic_type_info = type_info.generic_instantiation
            generic_convert_method = \
                self._generic_map[generic_type_info.generic_type]
            return generic_convert_method(generic_type_info, json_value)
        except KeyError:
            msg = ('Could not process the request, generic type '
                   '%s is not supported' % generic_type_info.generic_type)
            logger.exception(msg)
            raise werkzeug.exceptions.InternalServerError(msg)

    def visit(self, type_info, json_value):
        """
        Deserialize the given input using the metamodel type information

        :type  type_info:
            :class:`com.vmware.vapi.metadata.metamodel_client.Type`
        :param type_info: Metamodel type information
        :type  json_value: :class:`object`
        :param json_value: Value to be visited
        :rtype: :class:`vmware.vapi.data.value.DataValue`
        :return: DataValue created using the input
        """
        convert_method = self._category_map[type_info.category]
        return convert_method(type_info, json_value)

    def generate_operation_input(self, service_id, operation_id, input_data,
                                 mapping_type):
        """
        This method generates a StructValue corresponding to the Python dict
        (deserialized from JSON) suitable as an input value for the specified
        operation.

        :type  service_id: :class:`str`
        :param service_id: Identifier of the service to be invoked.
        :type  operation_id: :class:`str`
        :param operation_id: Identifier of the operation to be invoked.
        :type  input_data: :class:`dict`
        :param input_data: Dictionary object that represents the deserialized
            json input.
        :type  mapping_type:
            :class:`vmware.vapi.lib.constants.RestAnnotationType`
        :param mapping_type: Rest annotation type needed for handling map input
        :rtype: :class:`vmware.vapi.data.value.DataValue` or
        :return: DataValue created using the input
        """

        param_info_map = \
            self._metadata.service_map[service_id][operation_id].param_info_map

        self.mapping_type = mapping_type
        try:
            fields = {
                param_name: self.visit(param_info_map[str(param_name)].type,
                                       param_value)
                for param_name, param_value in six.iteritems(input_data)}
        except KeyError as e:
            msg = 'Unexpected parameter %s in JSON body' % e
            logger.exception(msg)
            raise werkzeug.exceptions.BadRequest(msg)
        except CoreException as e:
            msg = 'Unexpected input in JSON body: %s' % e
            logger.exception(msg)
            raise werkzeug.exceptions.BadRequest(msg)
        return StructValue(name=OPERATION_INPUT, values=fields)

    def map_uri_params(self, uri_parameters, service_id, operation_id):
        """
        If RequestMapping annotation is present, map URI parameters to
        respective parameter names. When request mapping is provided, it is
        possible that the uri parameter name is different from the canonical
        name of the parameter.

        :type  uri_parameters: :class:`dict` of :class:`str` and :class:`object`
        :param uri_parameters: Arguments parsed from the HTTP URL
        :type  service_id: :class:`str`
        :param service_id: Identifier of the service to be invoked
        :type  operation_id: :class:`str`
        :param operation_id: Identifier of the operation to be invoked
        :rtype: :class:`dict` of :class:`str` and :class:`object`
        :return: Arguments parsed from the HTTP URL - path_variable name
                 replaced with the canonical name of the parameter
        """
        service = self._metadata.service_map[service_id]
        operation = service[operation_id]

        if not operation.has_request_mapping_metadata():
            return uri_parameters

        path_variables_map = operation.path_variables_map

        # Update keys in uri_parameters from path variable name to canonical
        # name of the parameter
        return {path_variables_map[name]: value
                for name, value in six.iteritems(uri_parameters)}

    def _change_parameter_structure(self, output, names, value):
        """
        Multilevel parameter names are converted to dictionaries and the
        corresponding parameter values are set..

        For example, suppose the input parameter values are as follows:
            output = {'spec': 'index': 5}
            names = ['spec', 'body', 'user']
            value = 'Bob'

        Returns the result:
            {
                'spec': {
                    'index': 5,
                    'body': {
                        'user': 'Bob'
                    }
                }
            }

        :type  output: :class:`dict`
        :param output: Converted parameter dictionary.
        :type  names: :class: :`list`
        :param names: Multilevel parameter name list.
        :type  value: :class: :`string` or `list` or `dict`.
        :param value: Parameter value.
        :rtype: :class:`dict`
        :return: Parameter dictionary.
        """
        param_dict = dict(output)
        name = names.pop(0)
        if len(names) > 0:
            sub_param_dict = {}
            if name in output and isinstance(output[name], dict):
                sub_param_dict = output[name]
            param_dict[name] = self._change_parameter_structure(
                                    sub_param_dict, names, value)
        else:
            param_dict[name] = value
        return param_dict

    def map_query_params(self, request, service_id, operation_id):
        """
        For Verb annotation map Query parameters to respective parameter names.

        :type  service_id: :class:`str`
        :param service_id: Identifier of the service to be invoked
        :type  operation_id: :class:`str`
        :param operation_id: Identifier of the operation to be invoked
        :rtype: :class:`dict` of :class:`str` and :class:`object`
        :return: Query variables - canonical name of the parameter
        """
        service = self._metadata.service_map[service_id]
        operation = service[operation_id]

        if not operation.has_verb_metadata:
            return None

        output_data = {}
        for key, value in six.iteritems(
                        URLValueDeserializer.deserialize(request.query_string)):
            if key != TASK_REST_QUERY_PARAM and \
                                    key in operation.query_variables_map:
                output_data = self._change_parameter_structure(
                    output=output_data,
                    names=operation.query_variables_map[key].split('.'),
                    value=value
                )
        return output_data

    def map_header_params(self, request, service_id, operation_id):
        """
        For Verb annotation map Header parameters to respective parameter names.

        :type  service_id: :class:`str`
        :param service_id: Identifier of the service to be invoked
        :type  operation_id: :class:`str`
        :param operation_id: Identifier of the operation to be invoked
        :rtype: :class:`dict` of :class:`str` and :class:`object`
        :return: Header variables - canonical name of the parameter
        """
        service = self._metadata.service_map[service_id]
        operation = service[operation_id]

        if not operation.has_verb_metadata:
            return None

        output_data = {}
        for key, value in six.iteritems(request.headers):
            if key in operation.header_variables_map:
                output_data = self._change_parameter_structure(
                    output=output_data,
                    names=operation.header_variables_map[key].split('.'),
                    value=value
                )
        return output_data


class SecurityContextBuilder(object):
    """
    Helper class to build the appropriate security context based
    on the authentication information present in the request context.
    """
    def __init__(self, parsers):
        """
        Initialize SecurityContextBuilder

        :type  parsers: :class:`list` of :class:`str`
        :param parsers: List of names of Security Parsers
        """
        #
        # The order in the variable determines the order in which
        # the security parsers parse the request. The first parser
        # that can create a security context is used.
        #
        # If the request contains more than one kind of security
        # related information, the first one that gets parsed is used.
        # And if that information is invalid, 401 UNAUTHORIZED
        # is returned to the client. The other security information
        # that is not parsed is not taken into consideration.
        #
        # The user has to remove the invalid security context and invoke
        # the operation again.
        #
        self._parsers = [parser_class()
                         for parser_class in dynamic_import_list(parsers)]

    def build(self, request):
        """
        Build the security context based on the authentication information
        in the request context

        :type  request: :class:`werkzeug.wrappers.Request`
        :param request: Request object
        :rtype: :class:`vmware.vapi.core.SecurityContext`
        :return: Security context
        """
        for parser in self._parsers:
            security_context = parser.build(request)
            if security_context:
                return security_context
        return None


class ResponseCookieBuilder(object):
    """
    Helper class to build the appropriate cookies to be set
    in the response.
    """
    def __init__(self, session_methods):
        """
        Initialize ResponseCookieBuilder

        :type  session_methods: :class:`list` of :class:`str`
        :param session_methods: List of names of session login methods
        """
        self._session_methods = session_methods

    def build(self, service_id, operation_id, method_result):
        """
        Build the response cookie for the request

        :type  service_id: :class:`str`
        :param service_id: Identifier of the service to be invoked.
        :type  operation_id: :class:`str`
        :param operation_id: Identifier of the operation to be invoked.
        :type  method_result: :class:`vmware.vapi.core.MethodResult`
        :param method_result: MethodResult object to be serialized
        :rtype: :class:`dict` of :class:`str` and :class:`str`
        :return: Dictionary containing cookies that need to be set in the
            response.
        """
        session_method = '%s.%s' % (service_id, operation_id)
        if session_method in self._session_methods:
            if isinstance(method_result.output, (StringValue, SecretValue)):
                session_id = method_result.output.value
            return {REST_SESSION_ID_KEY: session_id}
        return None


class ApplicationContextBuilder(object):
    """
    Helper class to build the appropriate application context.
    """

    def build(self, request):
        """
        Build the application context based on the HTTP request. It will look
        for the header via which opID can be passed and put it into
        ApplicationContext or create the default one otherwise.

        :type  request: :class:`werkzeug.wrappers.Request`
        :param request: Request object
        :rtype: :class:`vmware.vapi.core.ApplicationContext`
        :return: Application context
        """
        headers = {}
        op_id = request.headers.get(REST_OP_ID_HEADER)
        if op_id:
            headers[OPID] = op_id

        locale = request.headers.get(HTTP_ACCEPT_LANGUAGE)
        if locale:
            headers[LOCALE] = locale

        locale_format = request.headers.get(HTTP_FORMAT_LOCALE)
        if locale_format:
            headers[HTTP_FORMAT_LOCALE] = locale_format

        timezone = request.headers.get(HTTP_TIMEZONE)
        if timezone:
            headers[HTTP_TIMEZONE] = timezone

        if headers:
            return ApplicationContext(headers)
        else:
            return create_default_application_context()


class RESTHandler(object):
    """
    Request handler that accept REST API calls and invoke the corresponding
    call on ApiProvider interface.
    """
    def __init__(self, api_provider, allow_cookies=True, provider_config=None):
        """
        Initialize RESTHandler

        :type  api_provider: :class:`vmware.vapi.core.ApiProvider`
        :param api_provider: ApiProvider to be used to invoke the vAPI requests
        :type  allow_cookies: :class:`bool`
        :param allow_cookies: Whether cookie support should be enabled
        :type  provider_config:
            :class:`vmware.vapi.settings.config.ProviderConfig` or :class:`None`
        :param provider_config: ApiProvider to be used to invoke the vAPI
            requests
        """
        self._api_provider = api_provider
        self.metadata = MetadataStore(api_provider)
        self.rest_rules = RoutingRuleGenerator(
            self.metadata, provider_config.get_rest_prefix()).rest_rules
        self.data_value_deserializer = DataValueDeserializer(self.metadata)
        self.application_context_builder = ApplicationContextBuilder()
        self.security_context_builder = SecurityContextBuilder(
            provider_config.get_rest_security_parsers())
        self.response_cookie_builder = ResponseCookieBuilder(
            provider_config.get_rest_session_methods()) \
            if allow_cookies else None

    def _get_input_value(self, service_id, operation_id, input_data,
                         mapping_type):
        """
        Generate the input DataValue for the given operation

        :type  service_id: :class:`str`
        :param service_id: Identifier of the service to be invoked.
        :type  operation_id: :class:`str`
        :param operation_id: Identifier of the operation to be invoked.
        :type  input_data: :class:`dict`
        :param input_data: Dictionary object that represents the deserialized
            json input.
        :type  mapping_type:
            :class:`vmware.vapi.lib.constants.RestAnnotationType`
        :param mapping_type: Rest annotation type needed for handling map input
        """
        try:
            return self.data_value_deserializer.generate_operation_input(
                service_id, operation_id, input_data, mapping_type)
        except werkzeug.exceptions.BadRequest as e:
            raise e
        except Exception as e:
            msg = 'Cannot process request: %s' % e
            logger.exception(msg)
            raise werkzeug.exceptions.InternalServerError(msg)

    def _serialize_output(
            self, request, service_id, operation_id, method_result,
            use_cookies, mapping_type):
        """
        Serialize the MethodResult object

        :type  request: :class:`werkzeug.wrappers.Request`
        :param request: Request object
        :type  service_id: :class:`str`
        :param service_id: Identifier of the service to be invoked.
        :type  operation_id: :class:`str`
        :param operation_id: Identifier of the operation to be invoked.
        :type  method_result: :class:`vmware.vapi.core.MethodResult`
        :param method_result: MethodResult object to be serialized
        :type  use_cookies: :class:`bool`
        :param use_cookies: Whether cookies are to be sent in response
        :rtype: :class:`tuple` of :class:`int`, :class:`str`, :class:`dict`
        :return: HTTP status code, serialized json output of the operation,
            a dictionary of response cookies and response http headers
        """
        status = None
        http_headers = None

        operation_summary = self.metadata.service_map[service_id][operation_id]
        response_headers = operation_summary.response_headers_map
        response_body_name = operation_summary.response_body_name

        if method_result.success():
            if self.response_cookie_builder is None or not use_cookies:
                cookies = None
            else:
                cookies = self.response_cookie_builder.build(
                    service_id, operation_id, method_result)
            output = method_result.output
            json_output = None
            if not isinstance(output, VoidValue):
                if not isinstance(output, StructValue):
                    # If the output is not a struct or error value,
                    # box it with {'value':..} to make it a
                    # good json output
                    output = StructValue(name='output', values={
                        'value': output
                    })

                new_rest = mapping_type == RestAnnotationType.VERB
                json_output = DataValueConverter.convert_to_json(output,
                                                                 new_rest)

            if operation_summary.success_response_code is not None:
                status = operation_summary.success_response_code
            else:
                status = 200
            if request.method == 'DELETE':
                status = 204
        else:
            error_name = method_result.error.name
            if error_name in self.metadata.all_errors_response_code_map:
                status = self.metadata.all_errors_response_code_map[error_name]

            cookies = None
            error = method_result.error
            new_rest = mapping_type == RestAnnotationType.VERB
            json_output = DataValueConverter.convert_to_json(error,
                                                             new_rest)
            if status is None:
                status = vapi_to_http_error_map.get(error.name, 500)

        if json_output:
            json_output_obj = json.loads(json_output)
            if response_headers:
                http_headers = {response_headers[key]: json_output_obj[key]
                                for key in json_output_obj.keys()
                                if key in response_headers}

            response_body = json.loads(json_output)
            if response_body_name:
                # if there is @Body annotation, the result can only contain
                # that field, nothing else.
                if response_body_name in response_body.keys():
                    json_output = json.dumps(response_body[response_body_name])
                else:
                    json_output = ''
            elif response_headers:
                for key in json_output_obj.keys():
                    if key in response_headers.keys():
                        del response_body[key]
                json_output = json.dumps(response_body)

        return (status, json_output, cookies, http_headers)

    def _find_matching_operation(
            self,
            request,
            service_id,
            uri_parameters,
            query_string_flat_dict,
            dispatch_info_list):
        """
        Find matching operation

        :type  request: :class:`werkzeug.wrappers.Request`
        :param request: Request object
        :type  service_id: :class:`str`
        :param service_id: Identifier of the service to be invoked.
        :type  uri_parameters: :class:`dict` of :class:`str` and :class:`object`
        :param uri_parameters: Arguments parsed from the HTTP URL
        :type  query_string_flat_dict: :class:`dict` of :class:`str` and :class:`object`  # pylint: disable=line-too-long
        :param query_string_flat_dict: Arguments parsed from the HTTP URL
        :type  dispatch_info_list:  :class:`list` of :class:`DispatchInfo`
        :param dispatch_info_list: List of dispatch info object
        :rtype: :class:`tuple` of :class:`str`, :class:`str`, :class:`dict`
        :return: Operation id, annotation type and response code map
        """

        operation_id = None
        mapping_type = RestAnnotationType.NONE
        arity = 0

        # If there's just one dispatch target get that operation id.
        if len(dispatch_info_list) == 1 and \
                dispatch_info_list[0].mapping_type != RestAnnotationType.VERB:
            operation_id = dispatch_info_list[0].operation_id
            mapping_type = dispatch_info_list[0].mapping_type
        else:
            for dispatch_info in dispatch_info_list:
                curr_op_id, curr_arity, curr_type = \
                    dispatch_info.get_operation_id(
                        request, query_string_flat_dict, uri_parameters)

                if curr_op_id is not None and curr_arity >= arity:
                    operation_id = curr_op_id
                    arity = curr_arity
                    mapping_type = curr_type

            if operation_id is None:
                action = query_string_flat_dict.get(
                    RestAnnotations.ACTION_PARAM)
                if action is not None:
                    msg = 'No matching method available for the requested ' + \
                        'action \'%s\' on the URL' % action
                else:
                    msg = 'No matching method available for the requested ' + \
                        'URL and HTTP method'
                logger.error(msg)
                raise werkzeug.exceptions.NotFound(msg)

        # Check if it's a task invocation
        is_task = query_string_flat_dict.get(TASK_REST_QUERY_PARAM)
        if is_task is not None:
            if is_task.lower() == 'true':
                operation_id = get_task_operation_name(operation_id)
            else:
                msg = 'Invalid value for "%s" query param' % (
                            TASK_REST_QUERY_PARAM)
                logger.error(msg)
                raise werkzeug.exceptions.BadRequest(msg)

        if operation_id not in self.data_value_deserializer.get_metadata().service_map[service_id]:  # pylint: disable=line-too-long
            msg = 'No matching method available for the requested ' + \
                  'URL and HTTP method'
            logger.error(msg)
            raise werkzeug.exceptions.NotFound(msg)

        return (operation_id, mapping_type)

    def _input_data_update(self, input_data, update_data):
        """
        The dictionary to merge.

        For example, if you have two dictionaries
        {
            'header_string': 'first',
            'query_bool': True,
            'spec': {
                'query_long': 1
            }
        }
        and
        {
            'body_double': 1.0,
            'spec': {
                'body_list': [2, 3, 4]
            }
        }
        the combined result
        output_data = {
            'header_string': 'first',
            'query_bool': True,
            'body_double': 1.0,
            'spec': {
                'query_long': 1,
                'body_list': [2, 3, 4]
            }
        }

        :type  input_data: :class:`dict`
        :param input_data: HTTP passes in the parameter dictionary.
        :type  update_data: :class:`dict`
        :param update_data: The parameter dictionary to be appended.
        :rtype :class:`dict`
        :return The updated HTTP passes in the parameter dictionary.
        """
        tmp = dict(input_data)
        for key, value in six.iteritems(update_data):
            if key in tmp and isinstance(tmp[key], dict) and \
                    isinstance(value, dict):
                tmp[key] = self._input_data_update(tmp[key], value)
            else:
                tmp.update({key: value})
        return tmp

    def invoke(self, request, endpoint, uri_parameters):
        """
        Handle the REST API invocation

        :type  request: :class:`werkzeug.wrappers.Request`
        :param request: Request object
        :type  endpoint: :class:`tuple` of :class:`tuple` of :class:`str` and
        :class:`str`
        :param endpoint: Tuple of service ID and dispatch_info is a tuples of
        tuples of (action, operation_id)
        :type  uri_parameters: :class:`dict` of :class:`str` and :class:`object`
        :param uri_parameters: Arguments parsed from the HTTP URL
        :rtype: :class:`tuple` of :class:`int`, :class:`str`, :class:`dict`
        :return: HTTP status code, serialized json output of the operation,
            response cookies and http headers
        """

        # Operation input is contributed by:
        # 1. URI parameters: Arguments parsed from HTTP URL
        # 2. Query parameters: Arguments parsed from the query string in the
        #               HTTP URL
        # 3. Header parameters: Arguments parsed from HTTP Header
        # 4. Request body: Arguments present in request body
        #
        # There is a definitive place for each argument, a particular argument
        # can be passed only using one of the above ways.
        #
        # Typically, main object identifiers are present in the base HTTP URL.
        # For GET operations, additional parameters can be provided in the query
        # string. For POST, PUT, PATCH, parameters can only be provided using
        # the body

        query_string_flat_dict = werkzeug.urls.url_decode(
            request.query_string)
        (service_id, dispatch_info_list) = endpoint
        (operation_id, mapping_type) = \
            self._find_matching_operation(request,
                                          service_id,
                                          uri_parameters,
                                          query_string_flat_dict,
                                          dispatch_info_list)

        input_data = self.data_value_deserializer.map_uri_params(
            uri_parameters, service_id, operation_id)

        if mapping_type == RestAnnotationType.VERB:
            input_data = self._input_data_update(
                input_data, self.data_value_deserializer.map_query_params(
                    request, service_id, operation_id))

            input_data = self._input_data_update(
                input_data, self.data_value_deserializer.map_header_params(
                    request, service_id, operation_id))

        if request.method == 'GET' and mapping_type != RestAnnotationType.VERB:
            query_string_parameters = URLValueDeserializer.deserialize(
                request.query_string, mapping_type)
            # Currently we put all query params for GET requests in request
            # body, don't do that for TASK_REST_QUERY_PARAM
            input_data = self._input_data_update(
                input_data, {
                    key: query_string_parameters[key]
                        for key in query_string_parameters
                            if key != TASK_REST_QUERY_PARAM})

        if request.method in ['PUT', 'PATCH', 'POST']:
            request_body_string = request.get_data()
            if request_body_string:
                try:
                    if isinstance(request_body_string, six.string_types):
                        input_data = self._input_data_update(
                            input_data, json.loads(
                                request_body_string,
                                parse_float=decimal.Decimal))
                    elif isinstance(request_body_string, six.binary_type):
                        input_data = self._input_data_update(
                            input_data, json.loads(
                                request_body_string.decode('utf-8'),
                                parse_float=decimal.Decimal))
                except ValueError as e:
                    msg = 'Invalid JSON in request body: %s' % e
                    logger.error(msg)
                    raise werkzeug.exceptions.BadRequest(msg)

        # Create request context
        security_context = self.security_context_builder.build(request)
        application_context = self.application_context_builder.build(request)
        ctx = ExecutionContext(
            application_context=application_context,
            security_context=security_context)

        # Set application context for the current operation/thread
        set_context(ctx)

        input_value = self._get_input_value(
            service_id, operation_id, input_data, mapping_type)

        method_result = self._api_provider.invoke(
            service_id, operation_id, input_value, ctx)

        # Clear application context from the current operation/thread
        clear_context()

        use_cookies = REQUIRE_HEADER_AUTHN not in request.headers
        return self._serialize_output(
            request, service_id, operation_id,
            method_result, use_cookies, mapping_type)



"""
Aggregator Api Provider
"""

__author__ = 'VMware, Inc.'
__copyright__ = 'Copyright 2015, 2017 VMware, Inc.  All rights reserved. -- VMware Confidential'  # pylint: disable=line-too-long


import requests
import traceback

from vmware.vapi.core import ApiProvider, MethodResult, ExecutionContext
from vmware.vapi.data.value import StructValue
from vmware.vapi.lib import connect
from vmware.vapi.lib.addr_url_parser import get_url_scheme
from vmware.vapi.lib.constants import (Introspection, OPERATION_INPUT)
from vmware.vapi.lib.log import get_vapi_logger
from vmware.vapi.lib.std import make_error_value_from_msgs, make_std_error_def
from vmware.vapi.l10n.runtime import message_factory
from vmware.vapi.provider import local as local_provider
from vmware.vapi.provider.local import LocalProvider
from vmware.vapi.provider.introspection import AggregatorIntrospector


# Configure logging
logger = get_vapi_logger(__name__)


class AggregatorProvider(ApiProvider):
    """
    AggregatorProvider is an aggregating implementation of the
    ApiProvider interface. It aggregates a bunch of ApiProvider
    instances and expose it.
    """
    def __init__(self):
        ApiProvider.__init__(self)
        self._name = 'ApiAggregator'

        # key = service_id, value = api interface
        self._service_id_map = {}

        # key : name, value : provider
        self._providers = {}

        # key : name, value : connection info
        self._provider_data = {}

        # List of all the services present in the local provider
        self._local_service_ids = []

        self._introspection_service_names = [
            Introspection.PROVIDER_SVC,
            Introspection.SERVICE_SVC,
            Introspection.OPERATION_SVC
        ]
        self._introspector = None
        self._authn_chain_processors = []

    _operation_not_found_def = make_std_error_def(
        'com.vmware.vapi.std.errors.operation_not_found')
    _internal_server_error_def = make_std_error_def(
        'com.vmware.vapi.std.errors.internal_server_error')
    _augmented_errors = (_operation_not_found_def,
                         _internal_server_error_def)

    ###########################################################################
    #
    # Support functions for aggregator services
    #
    ###########################################################################

    @staticmethod
    def get_service_identifiers(api_provider):
        """
        Invokes introspection service list on the Api Provider and retrieves
        the list of services

        :type  api_provider: :class:`vmware.vapi.core.ApiProvider`
        :param api_provider: ApiProvider instance to be used for retrieving
            service identifiers
        :rtype: :class:`list` of :class:`str`
        :return: List of service identifiers
        :raise: :class:`Exception`: if service identifiers could not be
            retrieved
        """
        ctx = ExecutionContext()
        service_name = Introspection.SERVICE_SVC
        operation_name = 'list'
        struct_value = StructValue(OPERATION_INPUT)
        method_result = api_provider.invoke(
            service_name, operation_name, struct_value, ctx)
        if method_result.success():
            return [service.value for service in method_result.output]
        else:
            raise Exception('Could not retrieve service identifiers: %s',
                            repr(method_result.error))

    def _process_child(self, provider_cfg, child):
        """
        Process a aggregator child provider_cfg

        :type  provider_cfg: :class:`dict`
        :param provider_cfg: Properties dictionary
        :type  child: :class:`str`
        :param child: Child to be processed
        """
        try:
            url = provider_cfg.get_child_url(child)
            connector = connect.get_requests_connector(
                session=requests.session(), url=url)
            api_provider = connector.get_api_provider()
            rpc_protocol = get_url_scheme(url)
            msg_protocol = provider_cfg.get_provider_message_format(child)
            if api_provider:
                self.register_provider(api_provider,
                                       child,
                                       rpc_protocol,
                                       msg_protocol,
                                       addr=url)
        except Exception as e:
            stack_trace = traceback.format_exc()
            raise Exception('Could not register %s at %s due to %s'
                            % (child, url, stack_trace))

    def register_by_properties(self, provider_cfg):
        """
        Register a provider using a provider_cfg dictionary

        :type  provider_cfg: :class:`dict`
        :param provider_cfg: Properties dictionary
        """
        self._name = provider_cfg.get_provider_name()
        if not self._name:
            self._name = 'ApiAggregator'

        #
        # We have three providers here:
        #   Aggregator: To aggregate services from local and remote providers
        #   IntrospectionProvider: To serve introspection requests. This has
        #       all the introspection services but it will decide whether to
        #       route the introspection call to LocalProvider or Remote provider
        #       based on where it is located.
        #   LocalProvider: To serve requests for services specified in
        #       'local.interfaces'. This also has introspection services that
        #       only provides introspection for services in this LocalProvider.
        #
        # For non-introspection calls:
        # - If service is in LocalProvider, flow is Aggregator -> LocalProvider
        # - If service is in remote ApiProvider, flow is Aggregator -> Remote
        #        api provider
        #
        # For introspection calls:
        # - For a service in remote ApiProvider, flow is
        # Aggregator -> Introspection services in IntrospectionProvider
        # - For a service in LocalProvider, flow is
        # Aggregator -> Introspection services in IntrospectionProvider
        # - For a service in introspection set of services, flow is
        # Aggregator -> IntrospectionProvider -> Introspection services in
        # LocalProvider (If it is handled by IntrospectionProvider, then we
        # will go into infinite recursive calls)
        #

        # Retrieve the local provider singleton instance to hold all the
        # services present in 'local.interfaces' property of in the properties
        # file. Even if there are no interfaces in local.interfaces, this is
        # required as it would serve introspection requests for introspection
        # api.
        l_provider = local_provider.get_provider()
        self._introspector = AggregatorIntrospector(
            self._name, l_provider)
        l_provider.register_by_properties(provider_cfg)

        # Create a local provider to hold the introspection services that
        # aggregates introspection information from the local provider and
        # all the remote providers.
        introspection_provider = LocalProvider(load_introspection=False)
        for service in self._introspector.get_introspection_services():
            # Register the introspection service with the local provider
            introspection_provider.add_interface(service)
            # Register the introspection service with this aggregator
            self.register_service(service.get_identifier().get_name(),
                                  introspection_provider)

        # Registering services from local provider
        service_ids = self.get_service_identifiers(l_provider)
        self._local_service_ids = service_ids
        for service_id in service_ids:
            if service_id not in self._introspection_service_names:
                self.register_service(service_id, l_provider)

        # Register the children
        for child in provider_cfg.get_child_providers():
            self._process_child(provider_cfg, child)

    def get_providers(self):
        """
        Return the connection information of ApiProviders
        registered with the AggregatorProvider

        :rtype: :class:`list` of :class:`tuple` of (
            :class:`str`, :class:`str`, :class:`str`)
        :return: Tuple containing rpc protocol, msg protocol and uri
        """
        return self._provider_data

    def register_service(self, service_id, provider):
        """
        Register an service with the AggregatorProvider

        :type  service_id: :class:`str`
        :param service_id: Service identifier
        :type  provider: :class:`vmware.vapi.core.ApiProvider`
        :param provider: ApiProvider impl. for the specified service identifier
        """
        if service_id in self._service_id_map:
            logger.error('Service already registered: %s', service_id)
        else:
            self._service_id_map[service_id] = provider
            self._introspector.add_service(service_id, provider)
            logger.info('Registering service: %s', service_id)

    def register_provider(self, provider, name, rpc, msg, addr):
        """
        Register a ApiProvider with the AggregatorProvider

        :type  provider: :class:`vmware.vapi.core.ApiProvider`
        :param provider: ApiProvider to be registered
        :type  name: :class:`str`
        :param name: Provider name
        :type  rpc: :class:`str`
        :param rpc: RPC Protocol
        :type  msg: :class:`str`
        :param msg: Msg Protocol
        :type  addr: :class:`str`
        :param addr: URI of the ApiProvider
        """
        self._provider_data[name] = (rpc, msg, addr)
        old_provider = self._providers.get(name)
        if old_provider:
            return

        service_ids = self.get_service_identifiers(provider)
        for service_id in service_ids:
            if service_id not in self._introspection_service_names:
                self.register_service(service_id, provider)
        self._providers[name] = provider

    def unregister_service(self, service_id):
        """
        Unregister an service from AggregatorProvider

        :type  service_id: :class:`str`
        :param service_id: service to be unregistered
        """
        if service_id in self._service_id_map:
            del self._service_id_map[service_id]
            self._introspector.remove_service(service_id)

    def unregister_provider(self, provider_name):
        """
        Unregister a provider from AggregatorProvider

        :type  provider_name: :class:`str`
        :param provider_name: Provider to be unregistered
        """
        provider = self._providers.get(provider_name)
        if not provider:
            return

        service_ids = self.get_service_identifiers(provider)
        for service_id in service_ids:
            self.unregister_service(service_id)
        del self._providers[provider_name]
        del self._provider_data[provider_name]

        logger.info('Unregistering Provider %s',
                    provider.get_definition())

    def unregister_provider_by_name(self, provider_name):
        """
        Unregister a provider from AggregatorProvider

        :type  provider_name: :class:`str`
        :param provider_name: Provider to be unregistered
        """
        self.unregister_provider(provider_name)

    def reload_providers(self):
        """
        Reload all the providers in the AggregatorProvider
        """
        providers = list(self._providers.values()) + [
            local_provider.get_provider()]
        for provider in providers:
            service_ids = self.get_service_identifiers(provider)
            for service_id in service_ids:
                self.register_service(service_id, provider)

    ###########################################################################
    #
    # Implementation of Api Provider interface
    #
    ###########################################################################

    def _invoke_int(self, service_id, operation_id, input_value, ctx):
        """
        Internal implementation of invoke method

        :type  service_id: :class:`str`
        :param service_id: Service identifier
        :type  operation_id: :class:`str`
        :param operation_id: Operation identifier
        :type  input_value: :class:`vmware.vapi.data.value.StructValue`
        :param input_value: Method input parameters
        :type  ctx: :class:`vmware.vapi.core.ExecutionContext`
        :param ctx: Execution context for this method

        :rtype: :class:`vmware.vapi.core.MethodResult`
        :return: Result of the method invocation
        """
        # Check if the provider exists
        provider = self._service_id_map.get(service_id)
        if not provider:
            msg = message_factory.get_message(
                'vapi.method.input.invalid.interface',
                service_id)
            logger.error(msg)
            error_value = make_error_value_from_msgs(
                self._operation_not_found_def, msg)
            return MethodResult(error=error_value)

        # Continue the authentication chain only if the target service
        # is not in the local provider of this process. i.e. for only
        # remote calls
        if service_id not in self._local_service_ids:
            for processor in self._authn_chain_processors:
                ctx.security_context = processor.next_context(
                    ctx.security_context)

        # Actual method execution
        try:
            method_result = provider.invoke(
                service_id, operation_id, input_value, ctx)
            return method_result
        except Exception as e:
            stack_trace = traceback.format_exc()
            logger.error(
                'Service: %s, Operation: %s, input_value %s: exception: %s',
                service_id, operation_id, input_value, stack_trace)
            msg = message_factory.get_message('vapi.method.invoke.exception',
                                              str(e))
            error_value = make_error_value_from_msgs(
                self._internal_server_error_def, msg)
            method_result = MethodResult(error=error_value)
            return method_result

    def invoke(self, service_id, operation_id, input_value, ctx):
        """
        Invokes the specified method using the execution context and
        the input provided

        :type  service_id: :class:`str`
        :param service_id: Service identifier
        :type  operation_id: :class:`str`
        :param operation_id: Operation identifier
        :type  input_value: :class:`vmware.vapi.data.value.StructValue`
        :param input_value: Method input parameters
        :type  ctx: :class:`vmware.vapi.core.ExecutionContext`
        :param ctx: Execution context for this method

        :rtype: :class:`vmware.vapi.core.MethodResult`
        :return: Result of the method invocation
        """
        logger.info("Started: Service: %s, Operation: %s, Ctx: %s",
                    service_id, operation_id, ctx)
        method_result = self._invoke_int(
            service_id, operation_id, input_value, ctx)
        logger.info("Finished: Service: %s. Operation %s, Ctx: %s",
                    service_id, operation_id, ctx)
        return method_result


# Single AggregatorProvider instance
_aggregator_provider = AggregatorProvider()


def get_provider():
    """
    Returns the singleton AggregatorProvider instance

    :rtype: :class:`vmware.vapi.provider.AggregatorProvider`
    :return: AggregatorProvider instance
    """
    return _aggregator_provider



# -*- coding: utf-8 -*-
#---------------------------------------------------------------------------
# Copyright 2023 VMware, Inc.  All rights reserved.

# AUTO GENERATED FILE -- DO NOT MODIFY!
#
# vAPI stub file for package com.vmware.esx.hcl.
#---------------------------------------------------------------------------

"""
The ``com.vmware.esx.hcl_client`` module provides classes to query the hardware
compatibility for an ESXi or a cluster.

"""

__author__ = 'VMware, Inc.'
__docformat__ = 'restructuredtext en'

import sys
from warnings import warn

from com.vmware.cis_client import Tasks
from vmware.vapi.stdlib.client.task import Task
from vmware.vapi.bindings import type
from vmware.vapi.bindings.converter import TypeConverter
from vmware.vapi.bindings.enum import Enum
from vmware.vapi.bindings.error import VapiError
from vmware.vapi.bindings.struct import VapiStruct
from vmware.vapi.bindings.stub import (
    ApiInterfaceStub, StubFactoryBase, VapiInterface)
from vmware.vapi.bindings.common import raise_core_exception
from vmware.vapi.data.validator import (UnionValidator, HasFieldsOfValidator)
from vmware.vapi.exception import CoreException
from vmware.vapi.lib.constants import TaskType
from vmware.vapi.lib.rest import OperationRestMetadata

class CompatibilityStatus(Enum):
    """
    The ``CompatibilityStatus`` class defines compatibility status of a given
    server or PCI device against a specific release of ESXi.

    .. note::
        This class represents an enumerated type in the interface language
        definition. The class contains class attributes which represent the
        values in the current version of the enumerated type. Newer versions of
        the enumerated type may contain new values. To use new values of the
        enumerated type in communication with a server that supports the newer
        version of the API, you instantiate this class. See :ref:`enumerated
        type description page <enumeration_description>`.
    """
    COMPATIBLE = None
    """
    When given hardware is certified for the specified ESXi release but no
    validation of the software of this hardware is performed.

    """
    INCOMPATIBLE = None
    """
    When given hardware is not certified for the specified ESXi release.

    """
    UNAVAILABLE = None
    """
    When there is no information about specified hardware.

    """
    CERTIFIED = None
    """
    When given hardware is certified for the specified ESXi release. Its
    software is also validated and it is also certified.

    """
    NOT_CERTIFIED = None
    """
    When given hardware is certified for the specified ESXi release. Its
    software is also validated and it is not certified.

    """

    def __init__(self, string):
        """
        :type  string: :class:`str`
        :param string: String value for the :class:`CompatibilityStatus` instance.
        """
        Enum.__init__(string)

CompatibilityStatus._set_values({
    'COMPATIBLE': CompatibilityStatus('COMPATIBLE'),
    'INCOMPATIBLE': CompatibilityStatus('INCOMPATIBLE'),
    'UNAVAILABLE': CompatibilityStatus('UNAVAILABLE'),
    'CERTIFIED': CompatibilityStatus('CERTIFIED'),
    'NOT_CERTIFIED': CompatibilityStatus('NOT_CERTIFIED'),
})
CompatibilityStatus._set_binding_type(type.EnumType(
    'com.vmware.esx.hcl.compatibility_status',
    CompatibilityStatus))




class Driver(VapiStruct):
    """
    The ``Driver`` class contains attributes describing information about a
    driver.

    .. tip::
        The arguments are used to initialize data attributes with the same
        names.
    """




    def __init__(self,
                 name=None,
                 version=None,
                 vendor=None,
                ):
        """
        :type  name: :class:`str`
        :param name: The name of the driver.
        :type  version: :class:`str`
        :param version: The version of the driver.
        :type  vendor: :class:`str` or ``None``
        :param vendor: The vendor that produced the driver.
            If None vendor is unknown.
        """
        self.name = name
        self.version = version
        self.vendor = vendor
        VapiStruct.__init__(self)


Driver._set_binding_type(type.StructType(
    'com.vmware.esx.hcl.driver', {
        'name': type.StringType(),
        'version': type.StringType(),
        'vendor': type.OptionalType(type.StringType()),
    },
    Driver,
    False,
    None))



class Firmware(VapiStruct):
    """
    The ``Firmware`` class contains attributes describing information about a
    firmware.

    .. tip::
        The arguments are used to initialize data attributes with the same
        names.
    """




    def __init__(self,
                 version=None,
                 vendor=None,
                ):
        """
        :type  version: :class:`str`
        :param version: The version of the firmware.
        :type  vendor: :class:`str` or ``None``
        :param vendor: The vendor that produced the firmware.
            If None vendor is unknown.
        """
        self.version = version
        self.vendor = vendor
        VapiStruct.__init__(self)


Firmware._set_binding_type(type.StructType(
    'com.vmware.esx.hcl.firmware', {
        'version': type.StringType(),
        'vendor': type.OptionalType(type.StringType()),
    },
    Firmware,
    False,
    None))



class PCIDevice(VapiStruct):
    """
    The ``PCIDevice`` class contains attributes describing information about a
    single PCI device on a host.

    .. tip::
        The arguments are used to initialize data attributes with the same
        names.
    """




    def __init__(self,
                 pci_address=None,
                 vid=None,
                 did=None,
                 svid=None,
                 ssid=None,
                 vendor=None,
                 model_name=None,
                 class_code=None,
                 sub_class_code=None,
                 driver=None,
                 firmware=None,
                 used_by_vsan=None,
                ):
        """
        :type  pci_address: :class:`str`
        :param pci_address: The device's address in a given ESXi host.
        :type  vid: :class:`str`
        :param vid: A unique number assigned to each computer hardware device that
            helps to identify the chipset manufacturer. For example, Dell,
            Broadcom, etc.
        :type  did: :class:`str`
        :param did: A unique number that identifies the specific device of the Vendor
            (VID).
        :type  svid: :class:`str`
        :param svid: A unique number that identifies the card manufacturer.
        :type  ssid: :class:`str`
        :param ssid: A unique number that identifies the specific device of Subsystem
            Vendor (SVID).
        :type  vendor: :class:`str`
        :param vendor: The name of the vendor.
        :type  model_name: :class:`str`
        :param model_name: The name of the device model.
        :type  class_code: :class:`str`
        :param class_code: Register that specifies the type of function the device performs.
        :type  sub_class_code: :class:`str`
        :param sub_class_code: Register that specifies the specific function the device performs.
        :type  driver: :class:`Driver`
        :param driver: Currently installed driver used by the device.
        :type  firmware: :class:`Firmware` or ``None``
        :param firmware: Currently installed firmware used by the device.
            If None firmware is unknown.
        :type  used_by_vsan: :class:`bool`
        :param used_by_vsan: Shows whether the device is part of VSAN cluster or not.
        """
        self.pci_address = pci_address
        self.vid = vid
        self.did = did
        self.svid = svid
        self.ssid = ssid
        self.vendor = vendor
        self.model_name = model_name
        self.class_code = class_code
        self.sub_class_code = sub_class_code
        self.driver = driver
        self.firmware = firmware
        self.used_by_vsan = used_by_vsan
        VapiStruct.__init__(self)


PCIDevice._set_binding_type(type.StructType(
    'com.vmware.esx.hcl.PCI_device', {
        'pci_address': type.StringType(),
        'vid': type.StringType(),
        'did': type.StringType(),
        'svid': type.StringType(),
        'ssid': type.StringType(),
        'vendor': type.StringType(),
        'model_name': type.StringType(),
        'class_code': type.StringType(),
        'sub_class_code': type.StringType(),
        'driver': type.ReferenceType(__name__, 'Driver'),
        'firmware': type.OptionalType(type.ReferenceType(__name__, 'Firmware')),
        'used_by_vsan': type.BooleanType(),
    },
    PCIDevice,
    False,
    None))



class Server(VapiStruct):
    """
    The ``Server`` class contains attributes describing information about a
    server.

    .. tip::
        The arguments are used to initialize data attributes with the same
        names.
    """




    def __init__(self,
                 model_name=None,
                 vendor=None,
                 cpu_series=None,
                 cpu_features=None,
                 bios=None,
                ):
        """
        :type  model_name: :class:`str`
        :param model_name: The name of the server model (for example, "PowerEdge R740xd").
        :type  vendor: :class:`str`
        :param vendor: The name of the vendor (for example, "Dell").
        :type  cpu_series: :class:`str`
        :param cpu_series: The CPU series name (for example, "Intel Xeon Gold 6100/5100,
            Silver 4100, Bronze 3100 (Skylake-SP) Series"). 
            
            **Note**: This attribute is initialized with the CPU *model* name;
            it's updated to the actual CPU series later, based on recognizing
            one of the CPU series the server hardware is certified with in the
            VCG.s
        :type  cpu_features: :class:`str`
        :param cpu_features: The current CPU features.
        :type  bios: :class:`Firmware`
        :param bios: Currently installed BIOS of the server.
        """
        self.model_name = model_name
        self.vendor = vendor
        self.cpu_series = cpu_series
        self.cpu_features = cpu_features
        self.bios = bios
        VapiStruct.__init__(self)


Server._set_binding_type(type.StructType(
    'com.vmware.esx.hcl.server', {
        'model_name': type.StringType(),
        'vendor': type.StringType(),
        'cpu_series': type.StringType(),
        'cpu_features': type.StringType(),
        'bios': type.ReferenceType(__name__, 'Firmware'),
    },
    Server,
    False,
    None))



class Notification(VapiStruct):
    """
    The ``Notification`` class contains attributes to describe any
    info/warning/error messages that Tasks can raise.

    .. tip::
        The arguments are used to initialize data attributes with the same
        names.
    """




    def __init__(self,
                 id=None,
                 time=None,
                 message=None,
                 resolution=None,
                ):
        """
        :type  id: :class:`str`
        :param id: The notification id.
        :type  time: :class:`datetime.datetime`
        :param time: The time the notification was raised/found.
        :type  message: :class:`com.vmware.vapi.std_client.LocalizableMessage`
        :param message: The notification message.
        :type  resolution: :class:`com.vmware.vapi.std_client.LocalizableMessage` or ``None``
        :param resolution: The resolution message, if any.
            Only :class:`set` if there is a resolution available for this
            notification.
        """
        self.id = id
        self.time = time
        self.message = message
        self.resolution = resolution
        VapiStruct.__init__(self)


Notification._set_binding_type(type.StructType(
    'com.vmware.esx.hcl.notification', {
        'id': type.StringType(),
        'time': type.DateTimeType(),
        'message': type.ReferenceType('com.vmware.vapi.std_client', 'LocalizableMessage'),
        'resolution': type.OptionalType(type.ReferenceType('com.vmware.vapi.std_client', 'LocalizableMessage')),
    },
    Notification,
    False,
    None))



class Notifications(VapiStruct):
    """
    The ``Notifications`` class contains info/warning/error messages that can
    be reported be the task.

    .. tip::
        The arguments are used to initialize data attributes with the same
        names.
    """




    def __init__(self,
                 info=None,
                 warnings=None,
                 errors=None,
                ):
        """
        :type  info: :class:`list` of :class:`Notification` or ``None``
        :param info: Info notification messages reported.
            Only :class:`set` if an info was reported by the task.
        :type  warnings: :class:`list` of :class:`Notification` or ``None``
        :param warnings: Warning notification messages reported.
            Only :class:`set` if an warning was reported by the task.
        :type  errors: :class:`list` of :class:`Notification` or ``None``
        :param errors: Error notification messages reported.
            Only :class:`set` if an error was reported by the task.
        """
        self.info = info
        self.warnings = warnings
        self.errors = errors
        VapiStruct.__init__(self)


Notifications._set_binding_type(type.StructType(
    'com.vmware.esx.hcl.notifications', {
        'info': type.OptionalType(type.ListType(type.ReferenceType(__name__, 'Notification'))),
        'warnings': type.OptionalType(type.ListType(type.ReferenceType(__name__, 'Notification'))),
        'errors': type.OptionalType(type.ListType(type.ReferenceType(__name__, 'Notification'))),
    },
    Notifications,
    False,
    None))



class CompatibilityData(VapiInterface):
    """
    This class provides methods to update the local compatibility data residing
    on the vCenter Appliance or to get information about the said data. The
    information in the data is generic VMware compatibility information for
    servers and devices.
    """

    _VAPI_SERVICE_ID = 'com.vmware.esx.hcl.compatibility_data'
    """
    Identifier of the service in canonical form.
    """
    def __init__(self, config):
        """
        :type  config: :class:`vmware.vapi.bindings.stub.StubConfiguration`
        :param config: Configuration to be used for creating the stub.
        """
        VapiInterface.__init__(self, config, _CompatibilityDataStub)
        self._VAPI_OPERATION_IDS = {}
        self._VAPI_OPERATION_IDS.update({'update_task': 'update$task'})

    class Status(VapiStruct):
        """
        The ``CompatibilityData.Status`` class contains attributes to describe the
        information available for the compatibility data.

        .. tip::
            The arguments are used to initialize data attributes with the same
            names.
        """




        def __init__(self,
                     updated_at=None,
                     notifications=None,
                    ):
            """
            :type  updated_at: :class:`datetime.datetime`
            :param updated_at: Indicates when the data was last updated.
            :type  notifications: :class:`Notifications`
            :param notifications: Notifications returned by the operation.
            """
            self.updated_at = updated_at
            self.notifications = notifications
            VapiStruct.__init__(self)


    Status._set_binding_type(type.StructType(
        'com.vmware.esx.hcl.compatibility_data.status', {
            'updated_at': type.DateTimeType(),
            'notifications': type.ReferenceType(__name__, 'Notifications'),
        },
        Status,
        False,
        None))



    def get(self):
        """
        Provides information about the compatibility data located on the
        vCenter Appliance.


        :rtype: :class:`CompatibilityData.Status`
        :return: Information about the compatibility data.
        :raise: :class:`com.vmware.vapi.std.errors_client.NotAllowedInCurrentState` 
            if there is no compatibility data on the vCenter executing the
            operation.
        :raise: :class:`com.vmware.vapi.std.errors_client.Unauthenticated` 
            if the caller is not authenticated.
        :raise: :class:`com.vmware.vapi.std.errors_client.ResourceInaccessible` 
            if the vCenter this API is executed on is not part of the Customer
            Experience Improvement Program (CEIP).
        :raise: :class:`com.vmware.vapi.std.errors_client.Error` 
            If there is some unknown error. The accompanying error message will
            give more details about the failure.
        :raise: :class:`com.vmware.vapi.std.errors_client.Unauthorized`
            if you do not have all of the privileges described as follows: 
            
            * Method execution requires
              ``VcIntegrity.HardwareCompatibility.Read``.
        """
        return self._invoke('get', None)


    def update_task(self):
        """
        Replaces the local compatibility data with the latest version found
        from VMware official source.


        :raise: :class:`com.vmware.vapi.std.errors_client.NotAllowedInCurrentState` 
            if there is compatibility data update in progress.
        :raise: :class:`com.vmware.vapi.std.errors_client.Unauthenticated` 
            if the caller is not authenticated.
        :raise: :class:`com.vmware.vapi.std.errors_client.ResourceInaccessible` 
            if the vCenter this API is executed on is not part of the Customer
            Experience Improvement Program (CEIP).
        :raise: :class:`com.vmware.vapi.std.errors_client.Error` 
            If there is some unknown error. The accompanying error message will
            give more details about the failure.
        """
        task_id = self._invoke('update$task',
                                None)
        task_svc = Tasks(self._config)
        task_instance = Task(task_id, task_svc, type.VoidType())
        return task_instance
class Reports(VapiInterface):
    """
    This class provides methods to download information generated from the
    hardware compatibility feature residing on the vCenter Appliance.
    """

    _VAPI_SERVICE_ID = 'com.vmware.esx.hcl.reports'
    """
    Identifier of the service in canonical form.
    """
    def __init__(self, config):
        """
        :type  config: :class:`vmware.vapi.bindings.stub.StubConfiguration`
        :param config: Configuration to be used for creating the stub.
        """
        VapiInterface.__init__(self, config, _ReportsStub)
        self._VAPI_OPERATION_IDS = {}

    class Token(VapiStruct):
        """
        The ``Reports.Token`` class contains information about the token required
        to be passed in the HTTP header in the HTTP GET request to generate the
        report.

        .. tip::
            The arguments are used to initialize data attributes with the same
            names.
        """




        def __init__(self,
                     token=None,
                     expiry=None,
                    ):
            """
            :type  token: :class:`str`
            :param token: A one-time, short-lived token required in the HTTP header of the
                request to the url. This token needs to be passed in as a header
                with the name "session-id".
            :type  expiry: :class:`datetime.datetime`
            :param expiry: Expiry time of the token
            """
            self.token = token
            self.expiry = expiry
            VapiStruct.__init__(self)


    Token._set_binding_type(type.StructType(
        'com.vmware.esx.hcl.reports.token', {
            'token': type.SecretType(),
            'expiry': type.DateTimeType(),
        },
        Token,
        False,
        None))


    class Location(VapiStruct):
        """
        The ``Reports.Location`` class contains the URI location to download
        generated compatibility report, as well as a token required (as a header on
        the HTTP GET request) to get the report. The validity of the token is 5
        minutes. After the token expires, any attempt to call the URI with said
        token will fail.

        .. tip::
            The arguments are used to initialize data attributes with the same
            names.
        """




        def __init__(self,
                     url=None,
                     report_token=None,
                    ):
            """
            :type  url: :class:`str`
            :param url: Compatibility report download URL.
            :type  report_token: :class:`Reports.Token`
            :param report_token: Information about the token required in the HTTP GET request to
                download the compatibility report.
            """
            self.url = url
            self.report_token = report_token
            VapiStruct.__init__(self)


    Location._set_binding_type(type.StructType(
        'com.vmware.esx.hcl.reports.location', {
            'url': type.URIType(),
            'report_token': type.ReferenceType(__name__, 'Reports.Token'),
        },
        Location,
        False,
        None))



    def get(self,
            report,
            ):
        """
        Returns the location :class:`Reports.Location` information for
        downloading a compatibility report.

        :type  report: :class:`str`
        :param report: identifier of hardware compatiblity report to be downloaded.
            The parameter must be an identifier for the resource type:
            ``com.vmware.esx.hcl.resources.CompatibilityReport``.
        :rtype: :class:`Reports.Location`
        :return: ``Reports.Location`` class which includes the URI to file, short
            lived token and expiry of the token in the
            :class:`Reports.Location` object.
        :raise: :class:`com.vmware.vapi.std.errors_client.NotFound` 
            if there is no report for the given id.
        :raise: :class:`com.vmware.vapi.std.errors_client.Unauthenticated` 
            if the caller is not authenticated.
        :raise: :class:`com.vmware.vapi.std.errors_client.ResourceInaccessible` 
            if the vCenter this API is executed on is not part of the Customer
            Experience Improvement Program (CEIP).
        :raise: :class:`com.vmware.vapi.std.errors_client.Error` 
            If there is some unknown error. The accompanying error message will
            give more details about the failure.
        :raise: :class:`com.vmware.vapi.std.errors_client.Unauthorized`
            if you do not have all of the privileges described as follows: 
            
            * Method execution requires
              ``VcIntegrity.HardwareCompatibility.Read``.
        """
        return self._invoke('get',
                            {
                            'report': report,
                            })
class _CompatibilityDataStub(ApiInterfaceStub):
    def __init__(self, config):
        # properties for get operation
        get_input_type = type.StructType('operation-input', {})
        get_error_dict = {
            'com.vmware.vapi.std.errors.not_allowed_in_current_state':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'NotAllowedInCurrentState'),
            'com.vmware.vapi.std.errors.unauthenticated':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'Unauthenticated'),
            'com.vmware.vapi.std.errors.resource_inaccessible':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'ResourceInaccessible'),
            'com.vmware.vapi.std.errors.error':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'Error'),

        }
        get_input_value_validator_list = [
        ]
        get_output_validator_list = [
        ]
        get_rest_metadata = OperationRestMetadata(
            http_method='GET',
            url_template='/esx/hcl/compatibility-data/status',
            path_variables={
            },
            query_parameters={
            },
            dispatch_parameters={
            },
            header_parameters={
            },
            dispatch_header_parameters={
            }
        )

        # properties for update operation
        update_input_type = type.StructType('operation-input', {})
        update_error_dict = {
            'com.vmware.vapi.std.errors.not_allowed_in_current_state':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'NotAllowedInCurrentState'),
            'com.vmware.vapi.std.errors.unauthenticated':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'Unauthenticated'),
            'com.vmware.vapi.std.errors.resource_inaccessible':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'ResourceInaccessible'),
            'com.vmware.vapi.std.errors.error':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'Error'),

        }
        update_input_value_validator_list = [
        ]
        update_output_validator_list = [
        ]
        update_rest_metadata = OperationRestMetadata(
            http_method='POST',
            url_template='/esx/hcl/compatibility-data',
            path_variables={
            },
            query_parameters={
            },
            dispatch_parameters={
                'action': 'download',
            },
            header_parameters={
            },
            dispatch_header_parameters={
            }
        )

        operations = {
            'get': {
                'input_type': get_input_type,
                'output_type': type.ReferenceType(__name__, 'CompatibilityData.Status'),
                'errors': get_error_dict,
                'input_value_validator_list': get_input_value_validator_list,
                'output_validator_list': get_output_validator_list,
                'task_type': TaskType.NONE,
            },
            'update$task': {
                'input_type': update_input_type,
                'output_type': type.IdType(resource_types='com.vmware.cis.TASK'),
                'errors': update_error_dict,
                'input_value_validator_list': update_input_value_validator_list,
                'output_validator_list': [],
                'task_type': TaskType.TASK_ONLY,
            },
        }
        rest_metadata = {
            'get': get_rest_metadata,
            'update': update_rest_metadata,
        }
        ApiInterfaceStub.__init__(
            self, iface_name='com.vmware.esx.hcl.compatibility_data',
            config=config, operations=operations, rest_metadata=rest_metadata,
            is_vapi_rest=True)

class _ReportsStub(ApiInterfaceStub):
    def __init__(self, config):
        # properties for get operation
        get_input_type = type.StructType('operation-input', {
            'report': type.IdType(resource_types='com.vmware.esx.hcl.resources.CompatibilityReport'),
        })
        get_error_dict = {
            'com.vmware.vapi.std.errors.not_found':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'NotFound'),
            'com.vmware.vapi.std.errors.unauthenticated':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'Unauthenticated'),
            'com.vmware.vapi.std.errors.resource_inaccessible':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'ResourceInaccessible'),
            'com.vmware.vapi.std.errors.error':
                type.ReferenceType('com.vmware.vapi.std.errors_client', 'Error'),

        }
        get_input_value_validator_list = [
        ]
        get_output_validator_list = [
        ]
        get_rest_metadata = OperationRestMetadata(
            http_method='GET',
            url_template='/esx/hcl/reports/{report}',
            path_variables={
                'report': 'report',
            },
            query_parameters={
            },
            dispatch_parameters={
            },
            header_parameters={
            },
            dispatch_header_parameters={
            }
        )

        operations = {
            'get': {
                'input_type': get_input_type,
                'output_type': type.ReferenceType(__name__, 'Reports.Location'),
                'errors': get_error_dict,
                'input_value_validator_list': get_input_value_validator_list,
                'output_validator_list': get_output_validator_list,
                'task_type': TaskType.NONE,
            },
        }
        rest_metadata = {
            'get': get_rest_metadata,
        }
        ApiInterfaceStub.__init__(
            self, iface_name='com.vmware.esx.hcl.reports',
            config=config, operations=operations, rest_metadata=rest_metadata,
            is_vapi_rest=True)


class StubFactory(StubFactoryBase):
    _attrs = {
        'CompatibilityData': CompatibilityData,
        'Reports': Reports,
        'hosts': 'com.vmware.esx.hcl.hosts_client.StubFactory',
    }

